#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import constants, utils
from servicecatalog_puppet.workflow.dependencies import tasks


class TerminateLaunchRoleConstraintsForSpokeLocalPortfolioTask(tasks.TaskWithReference):
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    portfolio = luigi.Parameter()
    spoke_local_portfolio_name = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_TASK

    @property
    def drift_token_parameters(self):
        return f"portfolio/{self.portfolio}"

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "spoke_local_portfolio_name": self.spoke_local_portfolio_name,
            "region": self.region,
            "account_id": self.account_id,
        }

    def run(self):
        stack_name = f"launch-constraints-for-{utils.slugify_for_cloudformation_stack_name(self.spoke_local_portfolio_name)}"

        with self.spoke_regional_client("cloudformation") as cloudformation:
            self.info(f"About to delete the stack: {stack_name}")
            cloudformation.ensure_deleted(StackName=stack_name)
        self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
from unittest import mock as mocker

import yaml

from servicecatalog_puppet import constants


@mocker.patch("servicecatalog_puppet.config.get_home_region")
@mocker.patch(
    "servicecatalog_puppet.config.betterboto_client.CrossAccountClientContextManager"
)
def test_get_config_without_a_default_region(
    mocked_betterboto_client, mocked_get_home_region
):
    # setup
    from servicecatalog_puppet import config as sut

    expected_result = {
        "Foo": "Bar",
    }
    fake_home_region = "eu-west-9"
    mocked_get_home_region.return_value = fake_home_region
    mocked_response = {"Parameter": {"Value": yaml.safe_dump(expected_result)}}
    mocked_betterboto_client.return_value.__enter__.return_value.get_parameter.return_value = (
        mocked_response
    )
    puppet_account_id = ""

    # exercise
    actual_result = sut.get_config(puppet_account_id)

    # verify
    assert actual_result == expected_result
    args, kwargs = mocked_betterboto_client.call_args
    assert "ssm" == args[0]
    assert {"region_name": fake_home_region} == kwargs


@mocker.patch(
    "servicecatalog_puppet.config.betterboto_client.CrossAccountClientContextManager"
)
def test_get_config_with_a_default_region(mocked_betterboto_client):
    # setup
    from servicecatalog_puppet import config as sut

    expected_result = {
        "Foo": "Bar",
    }
    fake_home_region = "eu-west-10"
    mocked_response = {"Parameter": {"Value": yaml.safe_dump(expected_result)}}
    mocked_betterboto_client.return_value.__enter__.return_value.get_parameter.return_value = (
        mocked_response
    )

    # exercise
    actual_result = sut.get_config(
        puppet_account_id="", default_region=fake_home_region
    )

    # verify
    assert actual_result == expected_result
    args, kwargs = mocked_betterboto_client.call_args
    assert "ssm" == args[0]
    assert {"region_name": fake_home_region} == kwargs


@mocker.patch(
    "servicecatalog_puppet.config.betterboto_client.CrossAccountClientContextManager"
)
def test_get_org_iam_role_arn(mocked_betterboto_client):
    # setup
    from servicecatalog_puppet import config as sut

    expected_result = "some_fake_arn"
    mocked_response = {"Parameter": {"Value": expected_result}}
    mocked_betterboto_client.return_value.__enter__.return_value.get_parameter.return_value = (
        mocked_response
    )
    mocked_get_home_region = mocker.patch.object(sut, "get_home_region")
    mocked_get_home_region.return_value = "us-east-9"
    puppet_account_id = ""

    # exercise
    actual_result = sut.get_org_iam_role_arn(puppet_account_id)

    # verify
    assert actual_result == expected_result
    args, kwargs = mocked_betterboto_client().__enter__().get_parameter.call_args
    assert args == ()
    assert kwargs == {"Name": constants.CONFIG_PARAM_NAME_ORG_IAM_ROLE_ARN}



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import logging

import networkx as nx

from servicecatalog_puppet import constants
from servicecatalog_puppet.waluigi.constants import (
    COMPLETED,
    ERRORED,
    NOT_SET,
    QUEUE_STATUS,
)


logger = logging.getLogger(constants.PUPPET_SCHEDULER_LOGGER_NAME)


def build_the_dag(tasks_to_run: dict):
    g = nx.DiGraph()
    # print("-- BUILDING THE DAG!!!")
    for uid, task in tasks_to_run.items():
        g.add_nodes_from(
            [
                (uid, task),
            ]
        )
        for duid in task.get("dependencies_by_reference", []):
            if tasks_to_run.get(duid):
                g.add_edge(uid, duid)
            else:
                logger.debug(
                    f"{duid} is not in the task reference - this is fine when running in spoke execution mode and when the task was executed within the hub"
                )

    for uid, task in tasks_to_run.items():
        if task.get(QUEUE_STATUS, NOT_SET) == COMPLETED:
            try:
                g.remove_node(uid)
            except nx.exception.NetworkXError as e:
                pass

        elif task.get(QUEUE_STATUS, NOT_SET) == ERRORED:
            # print(
            #     f"looking at task {uid} with status {task.get(QUEUE_STATUS, NOT_SET)}"
            # )
            for n in nx.ancestors(g, uid):
                try:
                    g.remove_node(n)
                except nx.exception.NetworkXError as e:
                    pass
            try:
                g.remove_node(uid)
            except nx.exception.NetworkXError as e:
                pass

    return g


def make_readable_in_codebuild_logs(input):
    numbers = "zero one two three four five six seven eight nine".split()
    numbers.extend("ten eleven twelve thirteen fourteen fifteen sixteen".split())
    numbers.extend("seventeen eighteen nineteen".split())
    numbers.extend(
        tens if ones == "zero" else (tens + "-" + ones)
        for tens in "twenty thirty forty fifty sixty seventy eighty ninety".split()
        for ones in numbers[0:10]
    )
    return numbers[input]



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
from servicecatalog_puppet import config, constants, task_reference_constants
from servicecatalog_puppet.commands.task_reference_helpers.generators import portfolios


def handle_launches(
    all_tasks,
    all_tasks_task_reference,
    item_name,
    puppet_account_id,
    section_name,
    task_reference,
    task_to_add,
):
    is_deploying_into_puppet_account = (
        task_to_add.get("account_id") == puppet_account_id
    )
    # GET THE HUB DETAILS TASK
    hub_portfolio_ref = f"{constants.PORTFOLIO_LOCAL}-{puppet_account_id}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
    if not all_tasks.get(hub_portfolio_ref):
        all_tasks[hub_portfolio_ref] = {
            "puppet_account_id": puppet_account_id,
            "task_reference": hub_portfolio_ref,
            "dependencies_by_reference": [],
            "account_id": puppet_account_id,
            "region": task_to_add.get("region"),
            "portfolio": task_to_add.get("portfolio"),
            "status": task_to_add.get("status"),
            "execution": constants.EXECUTION_MODE_HUB,
            "section_name": constants.PORTFOLIO_LOCAL,
            task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
            task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
            task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
        }
    all_tasks[hub_portfolio_ref][
        task_reference_constants.MANIFEST_SECTION_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
    all_tasks[hub_portfolio_ref][task_reference_constants.MANIFEST_ITEM_NAMES].update(
        task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES)
    )
    all_tasks[hub_portfolio_ref][task_reference_constants.MANIFEST_ACCOUNT_IDS].update(
        task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS)
    )

    spoke_portfolio_puppet_association_ref = None
    if is_deploying_into_puppet_account:
        portfolio_deploying_from = hub_portfolio_ref
    else:
        # share the portfolio and accept it
        sharing_mode = task_to_add.get(
            "sharing_mode", config.get_global_sharing_mode_default()
        )
        if sharing_mode == constants.SHARING_MODE_ACCOUNT:
            share_and_accept_ref = f"portfolio_share_and_accept-{task_to_add.get('account_id')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
            sharing_type = "ACCOUNT"
        elif sharing_mode == constants.SHARING_MODE_AWS_ORGANIZATIONS:
            share_and_accept_ref = f"portfolio_share_and_accept-{task_to_add.get('ou')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
            sharing_type = (
                "ORGANIZATIONAL_UNIT"
                if task_to_add.get("ou")[0:3] == "ou-"
                else "ORGANIZATION"
            )
        else:
            raise Exception(f"Unknown sharing mode: {sharing_mode}")

        describe_portfolio_shares_task_ref = (
            portfolios.get_or_create_describe_portfolio_shares_task_ref(
                all_tasks,
                puppet_account_id,
                sharing_type,
                hub_portfolio_ref,
                task_to_add,
            )
        )

        if not all_tasks.get(share_and_accept_ref):
            all_tasks[share_and_accept_ref] = {
                "puppet_account_id": puppet_account_id,
                "account_id": task_to_add.get("account_id"),
                "region": task_to_add.get("region"),
                "share_tag_options": task_to_add.get("share_tag_options"),
                "share_principals": task_to_add.get("share_principals"),
                "task_reference": share_and_accept_ref,
                "dependencies_by_reference": [
                    hub_portfolio_ref,
                    constants.CREATE_POLICIES,
                    describe_portfolio_shares_task_ref,
                ],
                "describe_portfolio_shares_task_ref": describe_portfolio_shares_task_ref,
                "portfolio": task_to_add.get("portfolio"),
                "execution": task_to_add.get("execution"),
                "portfolio_task_reference": hub_portfolio_ref,
                "section_name": f"portfolio-share-and-accept-{sharing_mode.lower()}",
                "ou_to_share_with": task_to_add.get("ou"),
                task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
            }
        all_tasks[share_and_accept_ref][
            task_reference_constants.MANIFEST_SECTION_NAMES
        ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
        all_tasks[share_and_accept_ref][
            task_reference_constants.MANIFEST_ITEM_NAMES
        ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
        all_tasks[share_and_accept_ref][
            task_reference_constants.MANIFEST_ACCOUNT_IDS
        ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))

        # GET THE IMPORTED PORTFOLIO
        spoke_portfolio_ref = f"{constants.PORTFOLIO_IMPORTED}-{task_to_add.get('account_id')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
        if not all_tasks.get(spoke_portfolio_ref):
            all_tasks[spoke_portfolio_ref] = {
                "puppet_account_id": puppet_account_id,
                "task_reference": spoke_portfolio_ref,
                "dependencies_by_reference": [share_and_accept_ref],
                "account_id": task_to_add.get("account_id"),
                "region": task_to_add.get("region"),
                "portfolio": task_to_add.get("portfolio"),
                "sharing_mode": task_to_add.get(
                    "sharing_mode", config.get_global_sharing_mode_default()
                ),
                "section_name": constants.PORTFOLIO_IMPORTED,
                task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
            }
        all_tasks[spoke_portfolio_ref][
            task_reference_constants.MANIFEST_SECTION_NAMES
        ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
        all_tasks[spoke_portfolio_ref][
            task_reference_constants.MANIFEST_ITEM_NAMES
        ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
        all_tasks[spoke_portfolio_ref][
            task_reference_constants.MANIFEST_ACCOUNT_IDS
        ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))
        portfolio_deploying_from = spoke_portfolio_ref

        spoke_portfolio_puppet_association_ref = f"{constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION}-{task_to_add.get('account_id')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
        if not all_tasks.get(spoke_portfolio_puppet_association_ref):
            all_tasks[spoke_portfolio_puppet_association_ref] = {
                "puppet_account_id": puppet_account_id,
                "task_reference": spoke_portfolio_puppet_association_ref,
                "portfolio_task_reference": spoke_portfolio_ref,
                "dependencies_by_reference": [spoke_portfolio_ref],
                "account_id": task_to_add.get("account_id"),
                "region": task_to_add.get("region"),
                "portfolio": task_to_add.get("portfolio"),
                "execution": task_to_add.get("execution"),
                "section_name": constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION,
                task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
            }
        all_tasks[spoke_portfolio_puppet_association_ref][
            task_reference_constants.MANIFEST_SECTION_NAMES
        ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
        all_tasks[spoke_portfolio_puppet_association_ref][
            task_reference_constants.MANIFEST_ITEM_NAMES
        ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
        all_tasks[spoke_portfolio_puppet_association_ref][
            task_reference_constants.MANIFEST_ACCOUNT_IDS
        ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))

        all_tasks[all_tasks_task_reference]["dependencies_by_reference"].append(
            spoke_portfolio_puppet_association_ref
        )

    # SETUP puppet association in hub so we can get the provisioning parameters
    hub_portfolio_puppet_association_ref = f"{constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION}-{task_to_add.get('puppet_account_id')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
    if not all_tasks.get(hub_portfolio_puppet_association_ref):
        all_tasks[hub_portfolio_puppet_association_ref] = {
            "puppet_account_id": puppet_account_id,
            "task_reference": hub_portfolio_puppet_association_ref,
            "portfolio_task_reference": hub_portfolio_ref,
            "dependencies_by_reference": [hub_portfolio_ref],
            "account_id": task_to_add.get("puppet_account_id"),
            "region": task_to_add.get("region"),
            "portfolio": task_to_add.get("portfolio"),
            "execution": task_to_add.get("execution"),
            "section_name": constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION,
            task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
            task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
            task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
        }
    all_tasks[hub_portfolio_puppet_association_ref][
        task_reference_constants.MANIFEST_SECTION_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
    all_tasks[hub_portfolio_puppet_association_ref][
        task_reference_constants.MANIFEST_ITEM_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
    all_tasks[hub_portfolio_puppet_association_ref][
        task_reference_constants.MANIFEST_ACCOUNT_IDS
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))

    # GET the provisioning parameters
    describe_provisioning_params_ref = f"{constants.DESCRIBE_PROVISIONING_PARAMETERS}-{puppet_account_id}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}-{task_to_add.get('product')}-{task_to_add.get('version')}"
    if not all_tasks.get(describe_provisioning_params_ref):
        all_tasks[describe_provisioning_params_ref] = {
            "puppet_account_id": puppet_account_id,
            "task_reference": describe_provisioning_params_ref,
            "dependencies_by_reference": [
                hub_portfolio_ref,
                hub_portfolio_puppet_association_ref,
            ],
            "account_id": puppet_account_id,
            "region": task_to_add.get("region"),
            "portfolio": task_to_add.get("portfolio"),
            "product": task_to_add.get("product"),
            "version": task_to_add.get("version"),
            "section_name": constants.DESCRIBE_PROVISIONING_PARAMETERS,
            task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
            task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
            task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
        }
    all_tasks[describe_provisioning_params_ref][
        task_reference_constants.MANIFEST_SECTION_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
    all_tasks[describe_provisioning_params_ref][
        task_reference_constants.MANIFEST_ITEM_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
    all_tasks[describe_provisioning_params_ref][
        task_reference_constants.MANIFEST_ACCOUNT_IDS
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))

    # GET all the products from the hub for the spoke
    portfolio_get_all_products_and_their_versions_ref = f"{constants.PORTFOLIO_GET_ALL_PRODUCTS_AND_THEIR_VERSIONS}-{puppet_account_id}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
    if not all_tasks.get(portfolio_get_all_products_and_their_versions_ref):
        all_tasks[portfolio_get_all_products_and_their_versions_ref] = {
            "execution": task_to_add.get("execution"),
            "puppet_account_id": puppet_account_id,
            "task_reference": portfolio_get_all_products_and_their_versions_ref,
            "dependencies_by_reference": [
                hub_portfolio_ref,
                hub_portfolio_puppet_association_ref,
            ],
            "portfolio_task_reference": hub_portfolio_ref,
            "account_id": puppet_account_id,
            "region": task_to_add.get("region"),
            "section_name": constants.PORTFOLIO_GET_ALL_PRODUCTS_AND_THEIR_VERSIONS,
            task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
            task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
            task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
        }
    all_tasks[portfolio_get_all_products_and_their_versions_ref][
        task_reference_constants.MANIFEST_SECTION_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
    all_tasks[portfolio_get_all_products_and_their_versions_ref][
        task_reference_constants.MANIFEST_ITEM_NAMES
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
    all_tasks[portfolio_get_all_products_and_their_versions_ref][
        task_reference_constants.MANIFEST_ACCOUNT_IDS
    ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))

    all_tasks[all_tasks_task_reference]["dependencies_by_reference"].append(
        portfolio_get_all_products_and_their_versions_ref
    )
    all_tasks[all_tasks_task_reference][
        "portfolio_get_all_products_and_their_versions_ref"
    ] = portfolio_get_all_products_and_their_versions_ref
    all_tasks[all_tasks_task_reference]["dependencies_by_reference"].append(
        describe_provisioning_params_ref
    )
    all_tasks[all_tasks_task_reference][
        "describe_provisioning_params_ref"
    ] = describe_provisioning_params_ref



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import click


def echo(message):
    click.secho(
        message,
    )


def warn(message):
    click.secho(
        message,
        err=True,
        fg="yellow",
    )


def error(message):
    click.secho(
        message,
        err=True,
        fg="red",
    )



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from servicecatalog_puppet.workflow import tasks_unit_tests_helper


class GeneratePoliciesTest(tasks_unit_tests_helper.PuppetTaskUnitTest):
    puppet_account_id = "puppet_account_id"
    region = "region"
    account_id = "account_id"
    organizations_to_share_with = dict(a=1)
    ous_to_share_with = dict(b=2)
    accounts_to_share_with = dict(c=3)

    def setUp(self) -> None:
        from servicecatalog_puppet.workflow.generate import generate_policies_task

        self.module = generate_policies_task

        self.sut = self.module.GeneratePolicies(
            **self.get_common_args(),
            account_id=self.account_id,
            region=self.region,
            organizations_to_share_with=self.organizations_to_share_with,
            ous_to_share_with=self.ous_to_share_with,
            accounts_to_share_with=self.accounts_to_share_with,
        )

        self.wire_up_mocks()

    def test_params_for_results_display(self):
        # setup
        expected_result = {
            "account_id": self.account_id,
            "region": self.region,
        }

        # exercise
        actual_result = self.sut.params_for_results_display()

        # verify
        self.assertEqual(expected_result, actual_result)



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import unittest

from servicecatalog_puppet import constants
from servicecatalog_puppet.waluigi.constants import (
    BLOCKED,
    COMPLETED,
    ERRORED,
    IN_PROGRESS,
    NOT_SET,
    QUEUE_STATUS,
)


task_to_run_reference = "task_to_run"
dependency_task_reference = "dependency"
non_related_task_reference = "non-related"


class TestTaskTopologicalGenerationsWithoutScheduler(unittest.TestCase):
    def setUp(self) -> None:
        self.maxDiff = None
        from servicecatalog_puppet.waluigi.shared_tasks import (
            task_topological_generations_without_scheduler,
        )

        self.sut = task_topological_generations_without_scheduler

    def generate_for_no_dependencies(self):
        task_to_run = dict(
            task_reference=task_to_run_reference, dependencies_by_reference=[]
        )
        all_tasks = dict()
        all_tasks[task_to_run_reference] = task_to_run

        return task_to_run, all_tasks, list(all_tasks.keys())

    def test_has_dependencies_remaining_with_no_dependencies(self):
        # setup
        task_to_run, all_tasks, _ = self.generate_for_no_dependencies()

        expected_is_currently_blocked = False
        expected_is_permanently_blocked = False

        # exercise
        (
            actual_is_currently_blocked,
            actual_is_permanently_blocked,
        ) = self.sut.has_dependencies_remaining(task_to_run, all_tasks)

        # verify
        self.assertEqual(expected_is_currently_blocked, actual_is_currently_blocked)
        self.assertEqual(expected_is_permanently_blocked, actual_is_permanently_blocked)

    def test_get_next_task_to_run_with_no_dependencies(self):
        # setup
        _, all_tasks, tasks_to_run = self.generate_for_no_dependencies()
        resources = dict()

        expected_next_task = all_tasks[task_to_run_reference]
        expected_should_shut_down = False

        # exercise
        actual_next_task, actual_should_shut_down = self.sut.get_next_task_to_run(
            tasks_to_run, resources, all_tasks
        )

        # verify
        self.assertEqual(expected_next_task, actual_next_task)
        self.assertEqual(expected_should_shut_down, actual_should_shut_down)

    def test_get_next_task_to_run_with_no_dependencies_and_task_complete(self):
        # setup
        _, all_tasks, tasks_to_run = self.generate_for_no_dependencies()
        resources = dict()
        all_tasks[task_to_run_reference][QUEUE_STATUS] = COMPLETED
        expected_next_task = None
        expected_should_shut_down = True

        # exercise
        actual_next_task, actual_should_shut_down = self.sut.get_next_task_to_run(
            tasks_to_run, resources, all_tasks
        )

        # verify
        self.assertEqual(expected_next_task, actual_next_task)
        self.assertEqual(expected_should_shut_down, actual_should_shut_down)

    def generate_has_dependencies_remaining_with_completed_dependencies(self):
        dependency_task = dict(
            task_reference=dependency_task_reference,
            dependencies_by_reference=[],
            QUEUE_STATUS=COMPLETED,
        )
        task_to_run = dict(
            task_reference=task_to_run_reference,
            dependencies_by_reference=[dependency_task_reference],
        )
        all_tasks = dict()
        all_tasks[task_to_run_reference] = task_to_run
        all_tasks[dependency_task_reference] = dependency_task

        return task_to_run, all_tasks, list(all_tasks.keys())

    def test_has_dependencies_remaining_with_completed_dependencies(self):
        # setup
        (
            task_to_run,
            all_tasks,
            _,
        ) = self.generate_has_dependencies_remaining_with_completed_dependencies()

        expected_is_currently_blocked = False
        expected_is_permanently_blocked = False

        # exercise
        (
            actual_is_currently_blocked,
            actual_is_permanently_blocked,
        ) = self.sut.has_dependencies_remaining(task_to_run, all_tasks)

        # verify
        self.assertEqual(expected_is_currently_blocked, actual_is_currently_blocked)
        self.assertEqual(expected_is_permanently_blocked, actual_is_permanently_blocked)

    def test_get_next_task_to_run_with_completed_dependencies(self):
        # setup
        (
            _,
            all_tasks,
            tasks_to_run,
        ) = self.generate_has_dependencies_remaining_with_completed_dependencies()

        resources = dict()
        expected_next_task_reference = all_tasks[task_to_run_reference]
        expected_should_shutdown = False

        # exercise
        actual_next_task, actual_should_shutdown = self.sut.get_next_task_to_run(
            tasks_to_run, resources, all_tasks
        )

        # verify
        self.assertEqual(expected_next_task_reference, actual_next_task)
        self.assertEqual(expected_should_shutdown, actual_should_shutdown)

    def generate_has_dependencies_remaining_with_not_set_dependencies(self):
        dependency_task = dict(
            task_reference=dependency_task_reference,
            dependencies_by_reference=[],
            QUEUE_STATUS=NOT_SET,
        )
        task_to_run = dict(
            task_reference=task_to_run_reference,
            dependencies_by_reference=[dependency_task_reference],
        )
        all_tasks = dict()
        all_tasks[task_to_run_reference] = task_to_run
        all_tasks[dependency_task_reference] = dependency_task

        return task_to_run, all_tasks, list(all_tasks.keys())

    def test_has_dependencies_remaining_with_not_set_dependencies(self):
        # setup
        (
            task_to_run,
            all_tasks,
            _,
        ) = self.generate_has_dependencies_remaining_with_not_set_dependencies()

        expected_is_currently_blocked = True
        expected_is_permanently_blocked = False

        # exercise
        (
            actual_is_currently_blocked,
            actual_is_permanently_blocked,
        ) = self.sut.has_dependencies_remaining(task_to_run, all_tasks)

        # verify
        self.assertEqual(expected_is_currently_blocked, actual_is_currently_blocked)
        self.assertEqual(expected_is_permanently_blocked, actual_is_permanently_blocked)

    def test_get_next_task_to_run_remaining_with_not_set_dependencies(self):
        # setup
        (
            _,
            all_tasks,
            tasks_to_run,
        ) = self.generate_has_dependencies_remaining_with_not_set_dependencies()

        resources = dict()
        expected_next_task = all_tasks[dependency_task_reference]
        expected_should_shut_down = False

        # exercise
        (
            actual_next_task,
            actual_should_shut_down,
        ) = self.sut.get_next_task_to_run(tasks_to_run, resources, all_tasks)

        # verify
        self.assertEqual(expected_next_task, actual_next_task)
        self.assertEqual(expected_should_shut_down, actual_should_shut_down)

    def generate_has_dependencies_remaining_with_failed_dependencies(self):
        dependency_task = dict(
            task_reference=dependency_task_reference,
            dependencies_by_reference=[],
            QUEUE_STATUS=ERRORED,
        )
        task_to_run = dict(
            task_reference=task_to_run_reference,
            dependencies_by_reference=[dependency_task_reference],
        )
        all_tasks = dict()
        all_tasks[task_to_run_reference] = task_to_run
        all_tasks[dependency_task_reference] = dependency_task

        return task_to_run, all_tasks, list(all_tasks.keys())

    def test_has_dependencies_remaining_with_failed_dependencies(self):
        # setup
        (
            task_to_run,
            all_tasks,
            _,
        ) = self.generate_has_dependencies_remaining_with_failed_dependencies()

        expected_is_currently_blocked = True
        expected_is_permanently_blocked = True

        # exercise
        (
            actual_is_currently_blocked,
            actual_is_permanently_blocked,
        ) = self.sut.has_dependencies_remaining(task_to_run, all_tasks)

        # verify
        self.assertEqual(expected_is_currently_blocked, actual_is_currently_blocked)
        self.assertEqual(expected_is_permanently_blocked, actual_is_permanently_blocked)

    def test_get_next_task_to_run_with_failed_dependencies(self):
        # setup
        (
            _,
            all_tasks,
            tasks_to_run,
        ) = self.generate_has_dependencies_remaining_with_failed_dependencies()

        resources = dict()
        expected_next_task = None
        expected_should_shutdown = True

        # exercise
        (
            actual_next_task,
            actual_should_shutdown,
        ) = self.sut.get_next_task_to_run(tasks_to_run, resources, all_tasks)

        # verify
        self.assertEqual(expected_next_task, actual_next_task)
        self.assertEqual(expected_should_shutdown, actual_should_shutdown)

    def generate_failed_dependencies_but_additional_tasks_are_present(self):
        all_tasks = dict()
        dependency_task = dict(
            task_reference=dependency_task_reference,
            dependencies_by_reference=[],
            QUEUE_STATUS=ERRORED,
        )
        all_tasks[dependency_task_reference] = dependency_task

        non_related_task = dict(
            task_reference=non_related_task_reference, dependencies_by_reference=[]
        )
        all_tasks[non_related_task_reference] = non_related_task

        task_to_run = dict(
            task_reference=task_to_run_reference,
            dependencies_by_reference=[dependency_task_reference],
        )
        all_tasks[task_to_run_reference] = task_to_run

        return task_to_run, all_tasks, list(all_tasks.keys())

    def test_has_dependencies_remaining_with_failed_dependencies_but_additional_tasks_are_present(
        self,
    ):
        # setup
        (
            task_to_run,
            all_tasks,
            _,
        ) = self.generate_failed_dependencies_but_additional_tasks_are_present()

        expected_is_currently_blocked = True
        expected_is_permanently_blocked = True

        # exercise
        (
            actual_is_currently_blocked,
            actual_is_permanently_blocked,
        ) = self.sut.has_dependencies_remaining(task_to_run, all_tasks)

        # verify
        self.assertEqual(expected_is_currently_blocked, actual_is_currently_blocked)
        self.assertEqual(expected_is_permanently_blocked, actual_is_permanently_blocked)

    def test_get_next_task_to_run_with_failed_dependencies_but_additional_tasks_are_present(
        self,
    ):
        # setup
        (
            _,
            all_tasks,
            tasks_to_run,
        ) = self.generate_failed_dependencies_but_additional_tasks_are_present()
        resources = dict()

        expected_next_task = all_tasks[non_related_task_reference]
        expected_should_shutdown = False

        # execute
        actual_next_task, actual_should_shutdown = self.sut.get_next_task_to_run(
            tasks_to_run, resources, all_tasks
        )

        # verify
        self.assertEqual(expected_should_shutdown, actual_should_shutdown)
        self.assertEqual(expected_next_task, actual_next_task)

    def generate_blocked_dependencies(self):
        all_tasks = dict()
        dependency_task = dict(
            task_reference=dependency_task_reference,
            dependencies_by_reference=[],
            QUEUE_STATUS=BLOCKED,
        )
        all_tasks[dependency_task_reference] = dependency_task

        task_to_run = dict(
            task_reference=task_to_run_reference,
            dependencies_by_reference=[dependency_task_reference],
        )
        all_tasks[task_to_run_reference] = task_to_run

        return task_to_run, all_tasks, list(all_tasks.keys())

    def test_has_dependencies_remaining_with_blocked_dependencies(
        self,
    ):
        # setup
        (
            task_to_run,
            all_tasks,
            _,
        ) = self.generate_blocked_dependencies()

        expected_is_currently_blocked = True
        expected_is_permanently_blocked = True

        # exercise
        (
            actual_is_currently_blocked,
            actual_is_permanently_blocked,
        ) = self.sut.has_dependencies_remaining(task_to_run, all_tasks)

        # verify
        self.assertEqual(expected_is_currently_blocked, actual_is_currently_blocked)
        self.assertEqual(expected_is_permanently_blocked, actual_is_permanently_blocked)

    def test_get_next_task_to_run_with_blocked_dependencies(
        self,
    ):
        # setup
        (
            _,
            all_tasks,
            tasks_to_run,
        ) = self.generate_blocked_dependencies()
        resources = dict()

        expected_next_task = None
        expected_should_shutdown = True

        # execute
        actual_next_task, actual_should_shutdown = self.sut.get_next_task_to_run(
            tasks_to_run, resources, all_tasks
        )

        # verify
        self.assertEqual(expected_should_shutdown, actual_should_shutdown)
        self.assertEqual(expected_next_task, actual_next_task)

    def generate_spoke_execution(self):
        all_tasks = {
            "create-policies": {
                "task_reference": "create-policies",
                "resources_required": [
                    "CLOUDFORMATION_CREATE_OR_UPDATE_eu-west-1_OF_XXXX"
                ],
                QUEUE_STATUS: COMPLETED,
            },
            constants.GENERATE_MANIFEST: {
                "task_reference": constants.GENERATE_MANIFEST,
                "dependencies_by_reference": ["create-policies"],
                QUEUE_STATUS: COMPLETED,
            },
            "run-deploy-in-spoke_YYYYY": {
                "task_reference": "run-deploy-in-spoke_YYYYY",
                "dependencies_by_reference": [constants.GENERATE_MANIFEST],
                QUEUE_STATUS: IN_PROGRESS,
            },
        }

        task_to_run = all_tasks.get("run-deploy-in-spoke_YYYYY")

        return task_to_run, all_tasks, list(all_tasks.keys())

    def test_has_dependencies_remaining_with_spoke_execution(
        self,
    ):
        # setup
        (
            task_to_run,
            all_tasks,
            _,
        ) = self.generate_spoke_execution()

        expected_is_currently_blocked = False
        expected_is_permanently_blocked = False

        # exercise
        (
            actual_is_currently_blocked,
            actual_is_permanently_blocked,
        ) = self.sut.has_dependencies_remaining(task_to_run, all_tasks)

        # verify
        self.assertEqual(expected_is_currently_blocked, actual_is_currently_blocked)
        self.assertEqual(expected_is_permanently_blocked, actual_is_permanently_blocked)

    def test_get_next_task_to_run_with_spoke_execution(
        self,
    ):
        # setup
        (
            _,
            all_tasks,
            tasks_to_run,
        ) = self.generate_spoke_execution()
        resources = dict()

        expected_next_task = None
        expected_should_shutdown = False

        # execute
        actual_next_task, actual_should_shutdown = self.sut.get_next_task_to_run(
            tasks_to_run, resources, all_tasks
        )

        # verify
        self.assertEqual(expected_should_shutdown, actual_should_shutdown)
        self.assertEqual(expected_next_task, actual_next_task)

    def generate_missing_generate_policies(self):
        all_tasks = {
            "non-specific-task": {
                "task_reference": "non-specific-task",
                "dependencies_by_reference": ["create-policies"],
            },
        }

        task_to_run = all_tasks.get("non-specific-task")

        return task_to_run, all_tasks, list(all_tasks.keys())

    def test_has_dependencies_remaining_with_missing_generate_policies(
        self,
    ):
        # setup
        (
            task_to_run,
            all_tasks,
            _,
        ) = self.generate_missing_generate_policies()

        expected_is_currently_blocked = False
        expected_is_permanently_blocked = False

        # exercise
        (
            actual_is_currently_blocked,
            actual_is_permanently_blocked,
        ) = self.sut.has_dependencies_remaining(task_to_run, all_tasks)

        # verify
        self.assertEqual(expected_is_currently_blocked, actual_is_currently_blocked)
        self.assertEqual(expected_is_permanently_blocked, actual_is_permanently_blocked)

    def test_get_next_task_to_run_with_missing_generate_policies(
        self,
    ):
        # setup
        (
            _,
            all_tasks,
            tasks_to_run,
        ) = self.generate_missing_generate_policies()
        resources = dict()

        expected_next_task = all_tasks.get("non-specific-task")
        expected_should_shutdown = False

        # execute
        actual_next_task, actual_should_shutdown = self.sut.get_next_task_to_run(
            tasks_to_run, resources, all_tasks
        )

        # verify
        self.assertEqual(expected_should_shutdown, actual_should_shutdown)
        self.assertEqual(expected_next_task, actual_next_task)



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import time
import traceback

from servicecatalog_puppet import serialisation_utils
from servicecatalog_puppet.serialisation_utils import unwrap
from servicecatalog_puppet.waluigi.constants import (
    BLOCKED,
    COMPLETED,
    CONTROL_EVENT__COMPLETE,
    ERRORED,
    IN_PROGRESS,
    NOT_SET,
    QUEUE_STATUS,
    RESOURCES_REQUIRED,
)
from servicecatalog_puppet.waluigi.dag_utils import logger
from servicecatalog_puppet.waluigi.locks.external import (
    are_resources_are_free_for_task_dict,
)
from servicecatalog_puppet.workflow.dependencies import task_factory


def has_dependencies_remaining(task_to_run, all_tasks):
    is_currently_blocked = False
    is_permanently_blocked = False
    for dependency in task_to_run.get("dependencies_by_reference"):
        if all_tasks.get(dependency):
            dependency_status = all_tasks[dependency].get(QUEUE_STATUS, NOT_SET)
            if dependency_status in [ERRORED, BLOCKED]:
                is_currently_blocked = is_permanently_blocked = True
                return is_currently_blocked, is_permanently_blocked
            else:
                if dependency_status != COMPLETED:
                    is_currently_blocked = True
                    return is_currently_blocked, is_permanently_blocked
    return is_currently_blocked, is_permanently_blocked


def get_next_task_to_run(tasks_to_run, resources, all_tasks):
    task_permanently_blocked_status = list()
    for task_reference_to_run in tasks_to_run:
        task_to_run = all_tasks[task_reference_to_run]
        status = task_to_run.get(QUEUE_STATUS, NOT_SET)
        # check if not running or has previously run
        if status == IN_PROGRESS:
            task_permanently_blocked_status.append(False)
        elif status == NOT_SET:
            is_currently_blocked, is_permanently_blocked = has_dependencies_remaining(
                task_to_run, all_tasks
            )
            if is_permanently_blocked:
                task_to_run[QUEUE_STATUS] = BLOCKED
                all_tasks[task_reference_to_run] = task_to_run
            task_permanently_blocked_status.append(is_permanently_blocked)
            if not is_currently_blocked:
                are_resources_are_free, _ = are_resources_are_free_for_task_dict(
                    task_to_run, resources
                )
                if are_resources_are_free:
                    return task_to_run, False
    return None, all(task_permanently_blocked_status)


def lock_next_task_to_run(next_task, resources, all_tasks):
    task_reference = next_task["task_reference"]
    next_task[QUEUE_STATUS] = IN_PROGRESS
    all_tasks[task_reference] = next_task
    for r in next_task.get(RESOURCES_REQUIRED, []):
        resources[r] = task_reference


def setup_next_task_to_run(lock, tasks_to_run, resources, all_tasks):
    with lock:
        next_task, can_shut_down = get_next_task_to_run(
            tasks_to_run, resources, all_tasks
        )
        if next_task is not None:
            lock_next_task_to_run(next_task, resources, all_tasks)
    return next_task, can_shut_down


def set_task_as_run(lock, next_task, all_tasks, resources, result):
    with lock:
        for r in next_task.get(RESOURCES_REQUIRED, []):
            try:
                del resources[r]
            except KeyError:
                logger.warn(
                    f"{next_task.get('task_reference')} tried to unlock {r} but it wasn't present"
                )
        next_task[QUEUE_STATUS] = result
        all_tasks[next_task["task_reference"]] = next_task


def worker_task(
    thread_name,
    lock,
    task_queue,
    results_queue,
    task_processing_time_queue,
    task_trace_queue,
    control_queue,
    control_event,
    manifest_files_path,
    manifest_task_reference_file_path,
    puppet_account_id,
    resources_file_path,
    all_tasks,
    resources,
    tasks_to_run,
):
    logger.info(f"starting up")
    should_run = True

    while should_run:
        next_task = None
        while next_task is None:
            next_task, can_shut_down = setup_next_task_to_run(
                lock, tasks_to_run, resources, all_tasks
            )
            if can_shut_down:
                if control_queue:
                    control_queue.put(CONTROL_EVENT__COMPLETE)
                if control_event:
                    control_event.set()
                return
            time.sleep(1)

        task_reference = next_task["task_reference"]
        task = task_factory.create(
            manifest_files_path=manifest_files_path,
            manifest_task_reference_file_path=manifest_task_reference_file_path,
            puppet_account_id=puppet_account_id,
            parameters_to_use=next_task,
        )
        logger.info(f"executing task: {task_reference}")
        task.on_task_start()
        start = time.time()
        task_type, task_details = task.get_processing_time_details()
        task_trace_queue.put(
            (start, task_type, task_details, True, thread_name),
        )
        try:
            task.execute()
        except Exception as e:
            end = time.time()
            duration = end - start
            result = ERRORED
            logger.error(f"executed task [failure]: {task_reference} failures: {e}")
            try:
                logger.error(f"---- START OF ERROR----")
                logger.error(f"Task {task_type}:")
                for l in serialisation_utils.dump(unwrap(task_details)).split("\n"):
                    logger.error(l)
                for l in traceback.format_exception(
                    e,
                    value=e,
                    tb=e.__traceback__,
                ):
                    for sl in l.split("\n"):
                        logger.error(f"{sl}")
                logger.error(f"---- END OF ERROR ----")
            except Exception as e2:
                logger.error(f"Exception raised: {e2}  whilst logging exection: {e}")
            task.on_task_failure(e, duration)
        else:
            end = time.time()
            duration = end - start
            result = COMPLETED
            logger.info(f"executed task [{result}]: {task_reference}")
            task.on_task_success(duration)

        task_processing_time_queue.put(
            (
                duration,
                task_type,
                task_details,
            ),
        )
        task_trace_queue.put(
            (end, task_type, task_details, False, thread_name),
        )

        # print(f"{pid} Worker {task_reference} waiting for lock to unlock resources", flush=True)
        logger.info(f"executed task [success]: {task_reference}")
        set_task_as_run(lock, next_task, all_tasks, resources, result)
        time.sleep(0.01)

    logger.info(f"shutting down")



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import multiprocessing
import os
import time

import networkx as nx

from servicecatalog_puppet.waluigi.constants import CONTROL_EVENT__COMPLETE
from servicecatalog_puppet.waluigi.dag_utils import (
    build_the_dag,
    logger,
    make_readable_in_codebuild_logs,
)
from servicecatalog_puppet.waluigi.shared_tasks import (
    task_processing_time,
    task_topological_generations_with_scheduler,
    task_topological_generations_without_scheduler,
    task_trace,
)
from servicecatalog_puppet.waluigi.shared_tasks.workers import (
    worker_requiring_scheduler,
)


QUEUE_REFILL_SLEEP_DURATION = 1


def get_tasks(scheduling_algorithm):
    if scheduling_algorithm == "topological_generations":
        return (
            worker_requiring_scheduler.worker_task,
            task_topological_generations_with_scheduler.scheduler_task,
        )
    if scheduling_algorithm == "topological_generations_without_scheduler":
        return task_topological_generations_without_scheduler.worker_task, None
    raise ValueError(f"Unsupported scheduling_algorithm: {scheduling_algorithm}")


def run(
    num_workers,
    tasks_reference,
    manifest_files_path,
    manifest_task_reference_file_path,
    puppet_account_id,
    execution_mode,
    scheduling_algorithm,
):
    logger.info(
        f"Executing {len(tasks_reference.keys())} tasks with {make_readable_in_codebuild_logs(num_workers)} processes in {execution_mode} with scheduling_algorithm {scheduling_algorithm}!"
    )

    manager = multiprocessing.Manager()
    all_tasks = manager.dict(tasks_reference)
    resources = manager.dict()
    tasks_to_run = manager.list()

    dag = build_the_dag(tasks_reference)
    generations = list(nx.topological_generations(dag))
    if not generations:
        raise ValueError("No tasks to run")
    while generations:
        tasks_to_run.extend(list(generations.pop()))

    resources_file_path = f"{manifest_files_path}/resources.json"
    start = time.time()
    os.environ["SCT_START_TIME"] = str(start)

    with open(resources_file_path, "w") as f:
        f.write("{}")

    QueueKlass = multiprocessing.Queue
    EventKlass = multiprocessing.Event
    ExecutorKlass = multiprocessing.Process
    LockKlass = multiprocessing.Lock

    lock = LockKlass()
    task_queue = QueueKlass()
    results_queue = QueueKlass()
    control_event = None
    task_processing_time_queue = QueueKlass()
    task_trace_queue = QueueKlass()
    control_queue = QueueKlass()
    complete_event = EventKlass()

    worker_task_args = (
        lock,
        task_queue,
        results_queue,
        task_processing_time_queue,
        task_trace_queue,
        control_queue,
        control_event,
        manifest_files_path,
        manifest_task_reference_file_path,
        puppet_account_id,
        resources_file_path,
        all_tasks,
        resources,
        tasks_to_run,
    )
    scheduler_task_args = (
        num_workers,
        task_queue,
        results_queue,
        control_queue,
        control_event,
        QUEUE_REFILL_SLEEP_DURATION,
        all_tasks,
        tasks_to_run,
    )
    task_processing_time_args = (
        task_processing_time_queue,
        complete_event,
        execution_mode,
    )
    task_trace_args = (
        task_trace_queue,
        complete_event,
        puppet_account_id,
        execution_mode,
    )

    worker_task_to_use, scheduler_task_to_use = get_tasks(scheduling_algorithm)

    processes = [
        ExecutorKlass(
            name=f"worker#{i}",
            target=worker_task_to_use,
            args=(str(i),) + worker_task_args,
        )
        for i in range(num_workers)
    ]
    scheduler_thread = None

    if scheduler_task_to_use:
        scheduler_thread = ExecutorKlass(
            name="scheduler",
            target=scheduler_task_to_use,
            args=scheduler_task_args,
        )
    on_task_processing_time_thread = ExecutorKlass(
        name="on_task_processing_time",
        target=task_processing_time.on_task_processing_time_task,
        args=task_processing_time_args,
    )
    on_task_trace_thread = ExecutorKlass(
        name="on_task_trace",
        target=task_trace.on_task_trace,
        args=task_trace_args,
    )

    on_task_processing_time_thread.start()
    on_task_trace_thread.start()

    for process in processes:
        process.start()
    if scheduler_thread:
        scheduler_thread.start()
    while True:
        logger.info("Waiting for shutdown message")
        message = control_queue.get()
        if message == CONTROL_EVENT__COMPLETE:
            logger.info(f"Got the {message}, starting shutdown process")
            break

    for process in processes:
        process.terminate()
    time.sleep(10)
    on_task_processing_time_thread.terminate()
    on_task_trace_thread.terminate()
    logger.info(f"Time taken = {time.time() - start:.10f}")



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import unittest
from unittest import mock


SPOKE_REGIONAL_CLIENT = "SPOKE_REGIONAL_CLIENT"


def mocked_client():
    context_handler_mock = mock.MagicMock()
    client_mock = mock.MagicMock()
    context_handler_mock.return_value.__enter__.return_value = client_mock
    return client_mock, context_handler_mock


class Paginator:
    def __init__(self, items, expected_kwargs):
        self.items = items
        self.expected_kwargs = expected_kwargs

    def paginate(self, **kwargs):
        if self.expected_kwargs != kwargs:
            raise Exception(
                f"Expected paginate kwargs {self.expected_kwargs} but got {kwargs}"
            )
        return self.items


class Manager:
    def __init__(self):
        self.expected_requests_and_responses = []
        self.index = 0

    def add_expected_request_and_response(
        self,
        client_type,
        client_name,
        uses_paginator,
        request_name,
        request_parameters,
        response_to_return,
    ):
        self.expected_requests_and_responses.append(
            (
                client_type,
                client_name,
                uses_paginator,
                request_name,
                request_parameters,
                response_to_return,
            )
        )

    def get_attribute_for(
        self,
        actual_client_type,
        actual_service,
        actual_method,
        actual_region_name,
        actual_retry_max_attempts,
    ):
        (
            expected_client_type,
            expected_client_name,
            expected_uses_paginator,
            expected_method,
            expected_parameters,
            response_to_return,
        ) = self.expected_requests_and_responses[self.index]
        # self.index += 1

        if actual_client_type != actual_client_type:
            raise Exception(
                f"Expected client type {expected_client_type} but got {actual_client_type}"
            )

        if actual_service != expected_client_name:
            raise Exception(
                f"Expected client name {expected_client_name} but got {actual_service}"
            )

        if expected_uses_paginator:
            if actual_method != "get_paginator":
                raise Exception(
                    f"Expected a paginator to be requested but {actual_method} was requested instead"
                )

            def paginator_to_return(*args, **kwargs):
                if args[0] != expected_method:
                    raise Exception(
                        f"Expected paginator: {expected_method} does not match actual: {args[0]}"
                    )
                return Paginator(response_to_return, expected_parameters)

            return paginator_to_return

        else:
            if actual_method != expected_method:
                raise Exception(
                    f"Expected request name {expected_method} but got {actual_method}"
                )

            def call_to_return(*args, **kwargs):
                if kwargs != expected_parameters:
                    raise Exception(
                        f"Expected parameters: {expected_parameters} does not match actual: {kwargs}"
                    )
                return response_to_return

            return call_to_return


class FakeClientContextManager:
    def __init__(
        self, client_type, manager, service, region_name=None, retry_max_attempts=None
    ):
        self.client_type = client_type
        self.manager = manager
        self.service = service
        self.region_name = region_name
        self.retry_max_attempts = retry_max_attempts
        self.expected_requests_and_responses = []

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.expected_requests_and_responses = []

    def __getattr__(self, attr):
        if attr == "manager":
            return self.manager
        elif attr == "client_type":
            return self.client_type
        elif attr == "args":
            return self.args
        elif attr == "kwargs":
            return self.kwargs
        else:
            return self.manager.get_attribute_for(
                SPOKE_REGIONAL_CLIENT,
                self.service,
                attr,
                self.region_name,
                self.retry_max_attempts,
            )


class PuppetTaskUnitTest(unittest.TestCase):
    task_reference = "task_reference"
    puppet_account_id = "puppet_account_id"
    manifest_file_path = "manifest_file_path"

    def get_common_args(self):
        return dict(
            task_reference="task_reference",
            manifest_task_reference_file_path="manifest_task_reference_file_path",
            manifest_files_path=self.manifest_file_path,
            dependencies_by_reference="dependencies_by_reference",
            puppet_account_id="puppet_account_id",
        )

    drift_token = "NOW"
    run_token = "NOW"

    def wire_up_mocks(self):
        self.sut.write_output = mock.MagicMock()
        self.sut.write_empty_output = mock.MagicMock()

        self.manager = Manager()

        # start of spoke_regional_client
        def fake_spoke_regional_client(
            service, region_name=None, retry_max_attempts=None
        ):
            return FakeClientContextManager(
                SPOKE_REGIONAL_CLIENT,
                self.manager,
                service,
                region_name,
                retry_max_attempts,
            )

        self.sut.spoke_regional_client = mock.MagicMock(
            side_effect=fake_spoke_regional_client
        )

    def add_expected_request_and_response(
        self,
        client_type,
        client_name,
        request_name,
        request_parameters,
        response_to_return,
    ):
        self.manager.add_expected_request_and_response(
            client_type,
            client_name,
            False,
            request_name,
            request_parameters,
            response_to_return,
        )

    def add_expected_paginated_request_and_response(
        self,
        client_type,
        client_name,
        request_name,
        request_parameters,
        response_to_return,
    ):
        self.manager.add_expected_request_and_response(
            client_type,
            client_name,
            True,
            request_name,
            request_parameters,
            response_to_return,
        )

    def assert_output(self, expected_output):
        self.sut.write_output.assert_called_once_with(expected_output)

    def assert_empty_output(self):
        self.sut.write_empty_output.assert_called_once()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi
import troposphere as t
from troposphere import iam

from servicecatalog_puppet import config, constants
from servicecatalog_puppet.workflow.dependencies import tasks


c7nTrailBucket = "c7nTrailBucket"


class ForwardEventsForAccountTask(tasks.TaskWithReferenceAndCommonParameters):
    c7n_account_id = luigi.Parameter()
    custodian_region = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
            "account_id": self.account_id,
        }

    def run(self):
        tpl = t.Template()
        tpl.description = (
            "event forwarder template for c7n created by service catalog puppet"
        )
        tpl.add_resource(
            iam.Role(
                "c7nEventForwarder",
                RoleName="c7nEventForwarder",
                Path="/servicecatalog-puppet/c7n/",
                Policies=[
                    iam.Policy(
                        PolicyName="AllowPutEvents",
                        PolicyDocument={
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Action": ["events:PutEvents"],
                                    "Resource": {
                                        "Fn::Sub": "arn:${AWS::Partition}:events:"
                                        + self.custodian_region
                                        + ":"
                                        + self.c7n_account_id
                                        + ":event-bus/default"
                                    },
                                    "Effect": "Allow",
                                }
                            ],
                        },
                    )
                ],
                AssumeRolePolicyDocument={
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Action": ["sts:AssumeRole"],
                            "Effect": "Allow",
                            "Principal": {"Service": ["events.amazonaws.com"]},
                        }
                    ],
                },
            )
        )

        template = tpl.to_yaml()
        with self.spoke_regional_client("cloudformation") as cloudformation:
            cloudformation.create_or_update(
                Capabilities=["CAPABILITY_NAMED_IAM"],
                ShouldUseChangeSets=False,
                StackName="servicecatalog-puppet-c7n-eventforwarding",
                TemplateBody=template,
                NotificationARNs=(
                    [
                        f"arn:{config.get_partition()}:sns:{self.region}:{self.puppet_account_id}:servicecatalog-puppet-cloudformation-regional-events"
                    ]
                    if self.should_use_sns
                    else []
                ),
                ShouldDeleteRollbackComplete=self.should_delete_rollback_complete_stacks,
                Tags=self.initialiser_stack_tags,
            )
        self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import copy

import luigi
import troposphere as t
import yaml
from troposphere import codebuild, s3

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks


class PrepareAccountForWorkspaceTask(tasks.TaskWithReference):
    account_id = luigi.Parameter()

    cachable_level = constants.CACHE_LEVEL_PERMANENT

    def params_for_results_display(self):
        return {
            "account_id": self.account_id,
        }

    def get_template(self):
        puppet_version = constants.VERSION
        description = f"""Bootstrap template used to configure spoke account for terraform use
                {{"version": "{puppet_version}", "framework": "servicecatalog-puppet", "role": "bootstrap-spoke-terraform"}}"""

        service_role = t.Sub(
            "arn:aws:iam::${AWS::AccountId}:role/servicecatalog-puppet/PuppetDeployInSpokeRole"
        )
        template = t.Template(Description=description)
        state = template.add_resource(
            s3.Bucket(
                "state",
                BucketName=t.Sub("sc-puppet-state-${AWS::AccountId}"),
                VersioningConfiguration=s3.VersioningConfiguration(Status="Enabled"),
                BucketEncryption=s3.BucketEncryption(
                    ServerSideEncryptionConfiguration=[
                        s3.ServerSideEncryptionRule(
                            ServerSideEncryptionByDefault=s3.ServerSideEncryptionByDefault(
                                SSEAlgorithm="AES256"
                            )
                        )
                    ]
                ),
                PublicAccessBlockConfiguration=s3.PublicAccessBlockConfiguration(
                    BlockPublicAcls=True,
                    BlockPublicPolicy=True,
                    IgnorePublicAcls=True,
                    RestrictPublicBuckets=True,
                ),
                Tags=t.Tags({"ServiceCatalogPuppet:Actor": "Framework"}),
            )
        )
        template.add_resource(
            s3.BucketPolicy(
                "statePolicy",
                Bucket=t.Ref(state),
                PolicyDocument={
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Action": [
                                "s3:GetObject*",
                                "s3:PutObject*",
                            ],
                            "Principal": {"AWS": self.puppet_account_id},
                            "Resource": t.Join("/", [t.GetAtt(state, "Arn"), "*"]),
                            "Effect": "Allow",
                            "Sid": "AllowPuppet",
                        },
                    ],
                },
            )
        )
        execute_build_spec = dict(
            version="0.2",
            phases=dict(
                install=dict(
                    commands=[
                        "mkdir -p /root/downloads",
                        "curl -s -qL -o /root/downloads/terraform_${TERRAFORM_VERSION}_linux_amd64.zip https://releases.hashicorp.com/terraform/${TERRAFORM_VERSION}/terraform_${TERRAFORM_VERSION}_linux_amd64.zip",
                        "unzip /root/downloads/terraform_${TERRAFORM_VERSION}_linux_amd64.zip -d /usr/bin/",
                        "chmod +x /usr/bin/terraform",
                        "terraform --version",
                        "aws s3 cp $ZIP source.zip",
                        "unzip source.zip",
                    ],
                ),
                pre_build=dict(
                    commands=[
                        "aws s3 cp $STATE_FILE terraform.tfstate || echo 'no statefile copied'",
                        'ASSUME_ROLE_ARN="arn:aws:iam::${TARGET_ACCOUNT}:role/servicecatalog-puppet/PuppetRole"',
                        "TEMP_ROLE=$(aws sts assume-role --role-arn $ASSUME_ROLE_ARN --role-session-name terraform)",
                        "export TEMP_ROLE",
                        'export AWS_ACCESS_KEY_ID=$(echo "${TEMP_ROLE}" | jq -r ".Credentials.AccessKeyId")',
                        'export AWS_SECRET_ACCESS_KEY=$(echo "${TEMP_ROLE}" | jq -r ".Credentials.SecretAccessKey")',
                        'export AWS_SESSION_TOKEN=$(echo "${TEMP_ROLE}" | jq -r ".Credentials.SessionToken")',
                        "aws sts get-caller-identity",
                        "terraform init",
                    ],
                ),
                build=dict(
                    commands=[
                        "terraform apply -auto-approve",
                    ]
                ),
                post_build=dict(
                    commands=[
                        "terraform output -json > outputs.json",
                        "unset AWS_ACCESS_KEY_ID",
                        "unset AWS_SECRET_ACCESS_KEY",
                        "unset AWS_SESSION_TOKEN",
                        "aws sts get-caller-identity",
                        "aws s3 cp terraform.tfstate $STATE_FILE",
                    ]
                ),
            ),
            artifacts=dict(
                files=[
                    "outputs.json",
                ],
            ),
        )
        execute_terraform = dict(
            Name=constants.EXECUTE_TERRAFORM_PROJECT_NAME,
            ServiceRole=service_role,
            Tags=t.Tags.from_dict(**{"ServiceCatalogPuppet:Actor": "Framework"}),
            Artifacts=codebuild.Artifacts(
                Type="S3",
                Location=t.Ref("state"),
                Path="terraform-executions",
                Name="artifacts-execute",
                NamespaceType="BUILD_ID",
            ),
            TimeoutInMinutes=480,
            Environment=codebuild.Environment(
                ComputeType="BUILD_GENERAL1_SMALL",
                Image=constants.CODEBUILD_DEFAULT_IMAGE,
                Type="LINUX_CONTAINER",
                EnvironmentVariables=[
                    codebuild.EnvironmentVariable(
                        Name="TERRAFORM_VERSION",
                        Type="PARAMETER_STORE",
                        Value=constants.DEFAULT_TERRAFORM_VERSION_PARAMETER_NAME,
                    ),
                ]
                + [
                    codebuild.EnvironmentVariable(
                        Name=name,
                        Type="PLAINTEXT",
                        Value="CHANGE_ME",
                    )
                    for name in ["TARGET_ACCOUNT", "ZIP", "STATE_FILE"]
                ],
            ),
            Source=codebuild.Source(
                BuildSpec=yaml.safe_dump(execute_build_spec),
                Type="NO_SOURCE",
            ),
            Description="Execute the given terraform in the given account using the given state file",
        )

        # execute
        template.add_resource(
            codebuild.Project("ExecuteTerraformProject", **execute_terraform)
        )

        # execute dry run
        execute_dry_run_terraform = copy.deepcopy(execute_terraform)
        execute_dry_run_terraform["Name"] = (
            constants.EXECUTE_DRY_RUN_TERRAFORM_PROJECT_NAME
        )
        execute_dry_run_terraform["Description"] = execute_dry_run_terraform[
            "Description"
        ].replace("Execute", "DRY RUN of Execute")
        execute_dry_run_build_spec = copy.deepcopy(execute_build_spec)
        execute_dry_run_build_spec["phases"]["build"]["commands"] = [
            "terraform plan -out=plan.bin",
            "terraform show -json plan.bin > plan.json",
        ]
        del execute_dry_run_build_spec["phases"]["post_build"]
        execute_dry_run_build_spec["artifacts"] = dict(
            files=[
                "plan.bin",
                "plan.json",
            ],
        )
        execute_dry_run_terraform["Source"] = codebuild.Source(
            BuildSpec=yaml.safe_dump(execute_dry_run_build_spec),
            Type="NO_SOURCE",
        )
        execute_dry_run_terraform["Artifacts"] = codebuild.Artifacts(
            Type="S3",
            Location=t.Ref("state"),
            Path="terraform-executions",
            Name="artifacts-execute-dry-run",
            NamespaceType="BUILD_ID",
        )
        template.add_resource(
            codebuild.Project(
                "ExecuteDryRunTerraformProject", **execute_dry_run_terraform
            )
        )

        # terminate
        terminate_terraform = copy.deepcopy(execute_terraform)
        terminate_terraform["Name"] = constants.TERMINATE_TERRAFORM_PROJECT_NAME
        terminate_terraform["Description"] = terminate_terraform["Description"].replace(
            "Execute", "Terminate"
        )
        terminate_build_spec = copy.deepcopy(execute_build_spec)
        terminate_build_spec["phases"]["build"]["commands"] = [
            "terraform destroy -auto-approve"
        ]
        terminate_build_spec["phases"]["post_build"]["commands"] = [
            "unset AWS_ACCESS_KEY_ID",
            "unset AWS_SECRET_ACCESS_KEY",
            "unset AWS_SESSION_TOKEN",
            "aws sts get-caller-identity",
            "aws s3 cp terraform.tfstate $STATE_FILE",
        ]
        del terminate_build_spec["artifacts"]
        terminate_terraform["Source"] = codebuild.Source(
            BuildSpec=yaml.safe_dump(terminate_build_spec),
            Type="NO_SOURCE",
        )
        terminate_terraform["Artifacts"] = codebuild.Artifacts(
            Type="S3",
            Location=t.Ref("state"),
            Path="terraform-executions",
            Name="artifacts-terminate",
            NamespaceType="BUILD_ID",
        )
        template.add_resource(
            codebuild.Project("TerminateTerraformProject", **terminate_terraform)
        )

        # terminate dry run
        termminate_dry_run_terraform = copy.deepcopy(execute_terraform)
        termminate_dry_run_terraform["Name"] = (
            constants.TERMINATE_DRY_RUN_TERRAFORM_PROJECT_NAME
        )
        new_description = termminate_dry_run_terraform["Description"].replace(
            "Execute", "DRY RUN of Terminate"
        )
        termminate_dry_run_terraform["Description"] = new_description
        termminate_dry_run_build_spec = copy.deepcopy(execute_build_spec)
        termminate_dry_run_build_spec["phases"]["build"]["commands"] = [
            "terraform plan -destroy -out=plan.bin",
            "terraform show -json plan.bin > plan.json",
        ]
        del termminate_dry_run_build_spec["phases"]["post_build"]
        termminate_dry_run_build_spec["artifacts"] = dict(
            files=[
                "plan.bin",
                "plan.json",
            ],
        )
        termminate_dry_run_terraform["Source"] = codebuild.Source(
            BuildSpec=yaml.safe_dump(termminate_dry_run_build_spec),
            Type="NO_SOURCE",
        )
        termminate_dry_run_terraform["Artifacts"] = codebuild.Artifacts(
            Type="S3",
            Location=t.Ref("state"),
            Path="terraform-executions",
            Name="artifacts-terminate-dry-run",
            NamespaceType="BUILD_ID",
        )
        template.add_resource(
            codebuild.Project(
                "TerminateDryRunTerraformProject", **termminate_dry_run_terraform
            )
        )
        return template.to_yaml()

    def run(self):
        template = self.get_template()
        with self.spoke_client("cloudformation") as cloudformation:
            cloudformation.create_or_update(
                StackName=constants.TERRAFORM_SPOKE_PREP_STACK_NAME,
                TemplateBody=template,
                ShouldDeleteRollbackComplete=self.should_delete_rollback_complete_stacks,
                Tags=self.initialiser_stack_tags,
            )

        self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
from servicecatalog_puppet import serialisation_utils


colors = dict()

default_color = "#88000022"


def generate_node(task_reference, task):
    reference = task_reference.replace("-", "_")
    section_name = task.get("section_name")
    account_id = task.get("account_id")
    region = task.get("region")
    return dict(
        color=colors.get(section_name, default_color),
        label=f"""<table border="0" cellborder="1" cellspacing="0" cellpadding="4">
			<tr> <td colspan='2'> <b>{task_reference}</b></td> </tr>
			<tr> <td align="left">section_name</td> <td>{section_name}</td> </tr>
			<tr> <td align="left">account_id</td> <td>{account_id}</td> </tr>
			<tr> <td align="left">region</td> <td>{region}</td> </tr>
		</table>""",
    )


def escape(input):
    return input.replace("-", "_").replace("/", "__")


def generate_edge(task_reference, dependency_by_reference):
    return f"{escape(task_reference)} -> {escape(dependency_by_reference)}"


def graph(content_file_path):
    content = open(content_file_path.name, "r").read()
    reference = serialisation_utils.load(content)
    all_tasks = reference.get("all_tasks")

    nodes = dict()
    edges = list()

    for task_reference, task in all_tasks.items():
        nodes[task_reference] = generate_node(task_reference, task)
        for dependency_by_reference in task.get("dependencies_by_reference", []):
            edges.append(generate_edge(task_reference, dependency_by_reference))

    nodes_as_a_graph = ""
    for task_reference, node in nodes.items():
        reference = escape(task_reference)
        nodes_as_a_graph += f"""{reference} [
            color="{node.get('color')}"
            label=<{node.get('label')}>
            shape=plain
        ]
        """

    for e in edges:
        nodes_as_a_graph += e + "\n"

    return f"""digraph workflow {{
        {nodes_as_a_graph}
    }}"""



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import os
import queue
import time

from betterboto import client as betterboto_client

from servicecatalog_puppet import (
    config,
    constants,
    environmental_variables,
    serialisation_utils,
)
from servicecatalog_puppet.commands import graph
from servicecatalog_puppet.serialisation_utils import unwrap
from servicecatalog_puppet.waluigi.dag_utils import logger


def on_task_trace(task_trace_queue, complete_event, puppet_account_id, execution_mode):
    bucket = f"sc-puppet-log-store-{puppet_account_id}"
    key_prefix = f"{os.getenv('CODEBUILD_BUILD_ID', f'local/{os.getenv(environmental_variables.DRIFT_TOKEN)}')}/traces"
    if execution_mode != constants.EXECUTION_MODE_SPOKE:
        with betterboto_client.CrossAccountClientContextManager(
            "s3",
            config.get_reporting_role_arn(config.get_executor_account_id()),
            "s3",
        ) as s3:
            while not complete_event.is_set():
                time.sleep(0.1)
                try:
                    (
                        t,
                        task_type,
                        task_params,
                        is_start,
                        thread_name,
                    ) = task_trace_queue.get(timeout=5)
                except queue.Empty:
                    continue
                else:
                    tz = (t - float(os.getenv("SCT_START_TIME", 0))) * 1000000
                    task_reference = task_params.get("task_reference")
                    s3.put_object(
                        Bucket=bucket,
                        Key=f"{key_prefix}/{tz}-{graph.escape(task_reference)}-{'start' if is_start else 'end'}.json",
                        Body=serialisation_utils.json_dumps(
                            {
                                "name": task_reference,
                                "cat": task_type,
                                "ph": "B" if is_start else "E",
                                "pid": 1,
                                "tid": thread_name,
                                "ts": tz,
                                "args": unwrap(task_params),
                            }
                        ),
                    )

        logger.info("shutting down")



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0


def test_codecommit_provider():
    # setup
    from . import bootstrap

    puppet_version = "0.10.0"
    all_regions = [
        "us-east-1",
        "us-east-2",
    ]
    repo_name = "TestRepoName"
    repo_branch = "mainly"
    source = dict(
        Provider="CodeCommit",
        Configuration=dict(RepositoryName=repo_name, BranchName=repo_branch),
    )
    is_caching_enabled = False
    is_manual_approvals = False
    scm_skip_creation_of_repo = False
    should_validate = False

    # exercise
    actual_result = bootstrap.get_template(
        puppet_version,
        all_regions,
        source,
        is_caching_enabled,
        is_manual_approvals,
        scm_skip_creation_of_repo,
        should_validate,
    )

    source_stage = actual_result.resources.get("Pipeline").Stages[0]
    action = source_stage.Actions[0]

    # verify
    assert source_stage.Name == "Source"
    assert action.ActionTypeId.Provider == "CodeCommit"
    assert actual_result.resources.get("CodeRepo").RepositoryName == repo_name
    assert action.Configuration.get("BranchName") == repo_branch



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import functools

from servicecatalog_puppet import constants, serialisation_utils


@functools.lru_cache(maxsize=5)
def get_m(manifest_file_path):
    with open(manifest_file_path, "r") as f:
        return serialisation_utils.load(f.read())


def create(
    manifest_files_path,
    manifest_task_reference_file_path,
    puppet_account_id,
    parameters_to_use,
):
    # TODO add in support for list launches and dry run
    section_name = parameters_to_use.get("section_name")
    minimum_common_parameters = dict(
        puppet_account_id=puppet_account_id,
        task_reference=parameters_to_use.get("task_reference"),
        manifest_task_reference_file_path=manifest_task_reference_file_path,
        manifest_files_path=manifest_files_path,
        dependencies_by_reference=parameters_to_use.get("dependencies_by_reference"),
    )
    common_parameters = dict(
        **minimum_common_parameters,
        account_id=parameters_to_use.get("account_id"),
        region=parameters_to_use.get("region"),
    )
    manifest_file_path = f"{manifest_files_path}/manifest-expanded.yaml"

    status = parameters_to_use.get("status")
    if section_name == constants.STACKS:
        if status == "terminated":
            from servicecatalog_puppet.workflow.stack.terminate_stack_task import (
                TerminateStackTask,
            )

            return TerminateStackTask(
                **common_parameters,
                stack_name=parameters_to_use.get("stack_name"),
                bucket=parameters_to_use.get("bucket"),
                key=parameters_to_use.get("key"),
                version_id=parameters_to_use.get("version_id"),
                launch_name=parameters_to_use.get("launch_name"),
                stack_set_name=parameters_to_use.get("stack_set_name"),
                capabilities=parameters_to_use.get("capabilities"),
                ssm_param_inputs=[],
                launch_parameters=parameters_to_use.get("launch_parameters"),
                manifest_parameters=parameters_to_use.get(""),
                account_parameters=parameters_to_use.get(""),
                retry_count=parameters_to_use.get("retry_count"),
                worker_timeout=parameters_to_use.get("worker_timeout"),
                ssm_param_outputs=[],
                requested_priority=parameters_to_use.get("requested_priority"),
                use_service_role=parameters_to_use.get("use_service_role"),
                execution=parameters_to_use.get("execution"),
            )
        else:
            from servicecatalog_puppet.workflow.stack.provision_stack_task import (
                ProvisionStackTask,
            )

            return ProvisionStackTask(
                **common_parameters,
                get_s3_template_ref=parameters_to_use.get("get_s3_template_ref"),
                stack_name=parameters_to_use.get("stack_name"),
                bucket=parameters_to_use.get("bucket"),
                key=parameters_to_use.get("key"),
                version_id=parameters_to_use.get("version_id"),
                launch_name=parameters_to_use.get("launch_name"),
                stack_set_name=parameters_to_use.get("stack_set_name"),
                capabilities=parameters_to_use.get("capabilities"),
                ssm_param_inputs=[],
                launch_parameters=parameters_to_use.get("launch_parameters"),
                manifest_parameters=parameters_to_use.get(""),
                account_parameters=parameters_to_use.get(""),
                retry_count=parameters_to_use.get("retry_count"),
                worker_timeout=parameters_to_use.get("worker_timeout"),
                ssm_param_outputs=[],
                requested_priority=parameters_to_use.get("requested_priority"),
                use_service_role=parameters_to_use.get("use_service_role"),
                execution=parameters_to_use.get("execution"),
                manifest_file_path=manifest_file_path,
                tags=parameters_to_use.get("tags"),
            )

    elif section_name == constants.LAUNCHES:
        if status == "terminated":
            from servicecatalog_puppet.workflow.launch.do_terminate_product_task import (
                DoTerminateProductTask,
            )

            return DoTerminateProductTask(
                **common_parameters,
                launch_name=parameters_to_use.get("launch_name"),
                portfolio=parameters_to_use.get("portfolio"),
                product=parameters_to_use.get("product"),
                version=parameters_to_use.get("version"),
                retry_count=parameters_to_use.get("retry_count"),
                worker_timeout=parameters_to_use.get("worker_timeout"),
                requested_priority=parameters_to_use.get("requested_priority"),
                execution=parameters_to_use.get("execution"),
            )
        else:
            from servicecatalog_puppet.workflow.launch.provision_product_task import (
                ProvisionProductTask,
            )

            return ProvisionProductTask(
                **common_parameters,
                launch_name=parameters_to_use.get("launch_name"),
                portfolio=parameters_to_use.get("portfolio"),
                product=parameters_to_use.get("product"),
                version=parameters_to_use.get("version"),
                portfolio_get_all_products_and_their_versions_ref=parameters_to_use.get(
                    "portfolio_get_all_products_and_their_versions_ref"
                ),
                describe_provisioning_params_ref=parameters_to_use.get(
                    "describe_provisioning_params_ref"
                ),
                retry_count=parameters_to_use.get("retry_count"),
                worker_timeout=parameters_to_use.get("worker_timeout"),
                requested_priority=parameters_to_use.get("requested_priority"),
                execution=parameters_to_use.get("execution"),
                tags=parameters_to_use.get("tags"),
                manifest_file_path=manifest_file_path,
            )

    elif section_name == constants.BOTO3_PARAMETERS:
        from servicecatalog_puppet.workflow.general import boto3_task

        return boto3_task.Boto3Task(
            **common_parameters,
            client=parameters_to_use.get("client"),
            use_paginator=parameters_to_use.get("use_paginator"),
            call=parameters_to_use.get("call"),
            arguments=parameters_to_use.get("arguments"),
            filter=parameters_to_use.get("filter"),
        )

    elif section_name == constants.SSM_PARAMETERS_WITH_A_PATH:
        from servicecatalog_puppet.workflow.ssm import get_ssm_parameter_task

        return get_ssm_parameter_task.GetSSMParameterByPathTask(
            **common_parameters,
            path=parameters_to_use.get("path"),
            jmespath_location=parameters_to_use.get("jmespath"),
        )

    elif section_name == constants.SSM_PARAMETERS:
        from servicecatalog_puppet.workflow.ssm import get_ssm_parameter_task

        return get_ssm_parameter_task.GetSSMParameterTask(
            **common_parameters,
            param_name=parameters_to_use.get("param_name"),
            jmespath_location=parameters_to_use.get("jmespath"),
        )

    elif section_name == constants.S3_GET_OBJECT:
        from servicecatalog_puppet.workflow.s3 import get_s3_object_task

        return get_s3_object_task.GetS3ObjectTask(
            **common_parameters,
            key=parameters_to_use.get("key"),
        )

    elif section_name == constants.SSM_OUTPUTS:
        from servicecatalog_puppet.workflow.ssm import ssm_outputs_task

        if parameters_to_use.get("status") == constants.TERMINATED:

            return ssm_outputs_task.TerminateSSMOutputsTasks(
                **common_parameters,
                param_name=parameters_to_use.get("param_name"),
            )

        else:
            return ssm_outputs_task.SSMOutputsTasks(
                **common_parameters,
                param_name=parameters_to_use.get("param_name"),
                stack_output=parameters_to_use.get("stack_output"),
                task_generating_output=parameters_to_use.get("task_generating_output"),
                task_generating_output_account_id=parameters_to_use.get(
                    "task_generating_output_account_id"
                ),
                task_generating_output_region=parameters_to_use.get(
                    "task_generating_output_region"
                ),
                task_generating_output_section_name=parameters_to_use.get(
                    "task_generating_output_section_name"
                ),
                task_generating_output_entity_name=parameters_to_use.get(
                    "task_generating_output_entity_name"
                ),
                task_generating_output_stack_set_name=parameters_to_use.get(
                    "task_generating_output_stack_set_name"
                ),
                task_generating_output_launch_name=parameters_to_use.get(
                    "task_generating_output_launch_name"
                ),
                force_operation=parameters_to_use.get("force_operation"),
            )
    elif section_name == constants.TAG_POLICIES:
        if status == "terminated":
            raise Exception(
                "No supported yet, raise a github issue if you would like to see this"
            )

        else:
            from servicecatalog_puppet.workflow.tag_policies import (
                do_execute_tag_policies_task,
            )

            return do_execute_tag_policies_task.DoExecuteTagPoliciesTask(
                **common_parameters,
                tag_policy_name=parameters_to_use.get("tag_policy_name"),
                get_or_create_policy_ref=parameters_to_use.get(
                    "get_or_create_policy_ref"
                ),
                ou_name=parameters_to_use.get("ou_name"),
                content=parameters_to_use.get("content"),
                description=parameters_to_use.get("description"),
                requested_priority=parameters_to_use.get("requested_priority"),
                manifest_file_path=manifest_file_path,
            )

    elif section_name == constants.SERVICE_CONTROL_POLICIES:
        if status == "terminated":
            from servicecatalog_puppet.workflow.service_control_policies import (
                do_terminate_service_control_policies_task,
            )

            return do_terminate_service_control_policies_task.DoTerminateServiceControlPoliciesTask(
                **common_parameters,
                service_control_policy_name=parameters_to_use.get(
                    "service_control_policy_name"
                ),
                get_or_create_policy_ref=parameters_to_use.get(
                    "get_or_create_policy_ref"
                ),
                ou_name=parameters_to_use.get("ou_name"),
                content=parameters_to_use.get("content"),
                description=parameters_to_use.get("description"),
                requested_priority=parameters_to_use.get("requested_priority"),
                manifest_file_path=manifest_file_path,
            )
        else:
            from servicecatalog_puppet.workflow.service_control_policies import (
                do_execute_service_control_policies_task,
            )

            return do_execute_service_control_policies_task.DoExecuteServiceControlPoliciesTask(
                **common_parameters,
                service_control_policy_name=parameters_to_use.get(
                    "service_control_policy_name"
                ),
                get_or_create_policy_ref=parameters_to_use.get(
                    "get_or_create_policy_ref"
                ),
                ou_name=parameters_to_use.get("ou_name"),
                content=parameters_to_use.get("content"),
                description=parameters_to_use.get("description"),
                requested_priority=parameters_to_use.get("requested_priority"),
                manifest_file_path=manifest_file_path,
            )

    elif section_name == constants.ASSERTIONS:
        from servicecatalog_puppet.workflow.assertions import do_assert_task

        return do_assert_task.DoAssertTask(
            **common_parameters,
            assertion_name=parameters_to_use.get("assertion_name"),
            execution=parameters_to_use.get("execution"),
            expected=parameters_to_use.get("expected"),
            actual=parameters_to_use.get("actual"),
            requested_priority=parameters_to_use.get("requested_priority"),
        )

    elif section_name == constants.SIMULATE_POLICIES:
        from servicecatalog_puppet.workflow.simulate_policies import (
            do_execute_simulate_policy_task,
        )

        return do_execute_simulate_policy_task.DoExecuteSimulatePolicyTask(
            **common_parameters,
            simulate_policy_name=parameters_to_use.get("simulate_policy_name"),
            execution=parameters_to_use.get("execution"),
            requested_priority=parameters_to_use.get("requested_priority"),
            simulation_type=parameters_to_use.get("simulation_type"),
            policy_source_arn=parameters_to_use.get("policy_source_arn"),
            policy_input_list=parameters_to_use.get("policy_input_list"),
            permissions_boundary_policy_input_list=parameters_to_use.get(
                "permissions_boundary_policy_input_list"
            ),
            action_names=parameters_to_use.get("action_names"),
            expected_decision=parameters_to_use.get("expected_decision"),
            resource_arns=parameters_to_use.get("resource_arns"),
            resource_policy=parameters_to_use.get("resource_policy"),
            resource_owner=parameters_to_use.get("resource_owner"),
            caller_arn=parameters_to_use.get("caller_arn"),
            context_entries=parameters_to_use.get("context_entries"),
            resource_handling_option=parameters_to_use.get("resource_handling_option"),
        )

    elif section_name == constants.LAMBDA_INVOCATIONS:
        from servicecatalog_puppet.workflow.lambda_invocations import (
            do_invoke_lambda_task,
        )

        return do_invoke_lambda_task.DoInvokeLambdaTask(
            **common_parameters,
            lambda_invocation_name=parameters_to_use.get("lambda_invocation_name"),
            function_name=parameters_to_use.get("function_name"),
            qualifier=parameters_to_use.get("qualifier"),
            invocation_type=parameters_to_use.get("invocation_type"),
            manifest_file_path=manifest_file_path,
        )

    elif section_name == constants.CODE_BUILD_RUNS:
        from servicecatalog_puppet.workflow.codebuild_runs import (
            do_execute_code_build_run_task,
        )

        return do_execute_code_build_run_task.DoExecuteCodeBuildRunTask(
            **common_parameters,
            code_build_run_name=parameters_to_use.get("code_build_run_name"),
            project_name=parameters_to_use.get("project_name"),
            manifest_file_path=manifest_file_path,
        )

    elif section_name == constants.SPOKE_LOCAL_PORTFOLIOS:
        if status == "terminated":
            from servicecatalog_puppet.workflow.portfolio.portfolio_management import (
                terminate_spoke_local_portfolio_task,
            )

            return (
                terminate_spoke_local_portfolio_task.TerminateSpokeLocalPortfolioTask(
                    **common_parameters,
                    portfolio=parameters_to_use.get("portfolio"),
                )
            )
        else:
            from servicecatalog_puppet.workflow.portfolio.portfolio_management import (
                create_spoke_local_portfolio_task,
            )

            return create_spoke_local_portfolio_task.CreateSpokeLocalPortfolioTask(
                **common_parameters,
                portfolio=parameters_to_use.get("portfolio"),
                portfolio_task_reference=parameters_to_use.get(
                    "portfolio_task_reference"
                ),
            )

    elif section_name == constants.PORTFOLIO_LOCAL:
        from servicecatalog_puppet.workflow.portfolio.portfolio_management import (
            get_portfolio_task,
        )

        return get_portfolio_task.GetPortfolioLocalTask(
            **common_parameters,
            portfolio=parameters_to_use.get("portfolio"),
            status=parameters_to_use.get("status"),
        )

    elif section_name == constants.PORTFOLIO_IMPORTED:
        from servicecatalog_puppet.workflow.portfolio.portfolio_management import (
            get_portfolio_task,
        )

        return get_portfolio_task.GetPortfolioImportedTask(
            **common_parameters,
            sharing_mode=parameters_to_use.get("sharing_mode"),
            portfolio=parameters_to_use.get("portfolio"),
        )

    elif section_name == constants.PORTFOLIO_ASSOCIATIONS:
        if status == "terminated":
            from servicecatalog_puppet.workflow.portfolio.associations import (
                terminate_associations_for_spoke_local_portfolio_task,
            )

            return terminate_associations_for_spoke_local_portfolio_task.TerminateAssociationsForSpokeLocalPortfolioTask(
                **common_parameters,
                portfolio=parameters_to_use.get("portfolio"),
                spoke_local_portfolio_name=parameters_to_use.get(
                    "spoke_local_portfolio_name"
                ),
            )
        else:
            from servicecatalog_puppet.workflow.portfolio.associations import (
                create_associations_for_spoke_local_portfolio_task,
            )

            return create_associations_for_spoke_local_portfolio_task.CreateAssociationsForSpokeLocalPortfolioTask(
                **common_parameters,
                spoke_local_portfolio_name=parameters_to_use.get(
                    "spoke_local_portfolio_name"
                ),
                associations=parameters_to_use.get("associations"),
                portfolio=parameters_to_use.get("portfolio"),
                portfolio_task_reference=parameters_to_use.get(
                    "portfolio_task_reference"
                ),
            )

    elif section_name == constants.PORTFOLIO_CONSTRAINTS_LAUNCH:
        if status == "terminated":
            from servicecatalog_puppet.workflow.portfolio.constraints_management import (
                terminate_launch_role_constraints_for_spoke_local_portfolio_task,
            )

            return terminate_launch_role_constraints_for_spoke_local_portfolio_task.TerminateLaunchRoleConstraintsForSpokeLocalPortfolioTask(
                **common_parameters,
                portfolio=parameters_to_use.get("portfolio"),
                spoke_local_portfolio_name=parameters_to_use.get(
                    "spoke_local_portfolio_name"
                ),
            )

        else:
            from servicecatalog_puppet.workflow.portfolio.constraints_management import (
                create_launch_role_constraints_for_spoke_local_portfolio_task,
            )

            return create_launch_role_constraints_for_spoke_local_portfolio_task.CreateLaunchRoleConstraintsForSpokeLocalPortfolioTask(
                **common_parameters,
                spoke_local_portfolio_name=parameters_to_use.get(
                    "spoke_local_portfolio_name"
                ),
                launch_constraints=parameters_to_use.get("launch_constraints"),
                portfolio=parameters_to_use.get("portfolio"),
                portfolio_task_reference=parameters_to_use.get(
                    "portfolio_task_reference"
                ),
                portfolio_get_all_products_and_their_versions_ref=parameters_to_use.get(
                    "portfolio_get_all_products_and_their_versions_ref"
                ),
            )

    elif section_name == constants.PORTFOLIO_CONSTRAINTS_RESOURCE_UPDATE:
        if status == "terminated":
            from servicecatalog_puppet.workflow.portfolio.constraints_management import (
                terminate_resource_update_constraints_for_spoke_local_portfolio_task,
            )

            return terminate_resource_update_constraints_for_spoke_local_portfolio_task.TerminateResourceUpdateConstraintsForSpokeLocalPortfolioTask(
                **common_parameters,
                portfolio=parameters_to_use.get("portfolio"),
                spoke_local_portfolio_name=parameters_to_use.get(
                    "spoke_local_portfolio_name"
                ),
            )
        else:
            from servicecatalog_puppet.workflow.portfolio.constraints_management import (
                create_resource_update_constraints_for_spoke_local_portfolio_task,
            )

            return create_resource_update_constraints_for_spoke_local_portfolio_task.CreateUpdateResourceConstraintsForSpokeLocalPortfolioTask(
                **common_parameters,
                spoke_local_portfolio_name=parameters_to_use.get(
                    "spoke_local_portfolio_name"
                ),
                resource_update_constraints=parameters_to_use.get(
                    "resource_update_constraints"
                ),
                portfolio=parameters_to_use.get("portfolio"),
                portfolio_task_reference=parameters_to_use.get(
                    "portfolio_task_reference"
                ),
                portfolio_get_all_products_and_their_versions_ref=parameters_to_use.get(
                    "portfolio_get_all_products_and_their_versions_ref"
                ),
            )

    elif section_name == constants.PORTFOLIO_COPY:
        if status == "terminated":
            raise Exception("Not supported yet")
        else:
            from servicecatalog_puppet.workflow.portfolio.portfolio_management import (
                copy_into_spoke_local_portfolio_task,
            )

            return copy_into_spoke_local_portfolio_task.CopyIntoSpokeLocalPortfolioTask(
                **common_parameters,
                portfolio_task_reference=parameters_to_use.get(
                    "portfolio_task_reference"
                ),
                portfolio_get_all_products_and_their_versions_ref=parameters_to_use.get(
                    "portfolio_get_all_products_and_their_versions_ref"
                ),
                portfolio_get_all_products_and_their_versions_for_hub_ref=parameters_to_use.get(
                    "portfolio_get_all_products_and_their_versions_for_hub_ref"
                ),
            )

    elif section_name == constants.PORTFOLIO_IMPORT:
        if status == "terminated":
            raise Exception("Not supported yet")
        else:
            from servicecatalog_puppet.workflow.portfolio.portfolio_management import (
                import_into_spoke_local_portfolio_task,
            )

            return import_into_spoke_local_portfolio_task.ImportIntoSpokeLocalPortfolioTask(
                **common_parameters,
                portfolio_task_reference=parameters_to_use.get(
                    "portfolio_task_reference"
                ),
                hub_portfolio_task_reference=parameters_to_use.get(
                    "hub_portfolio_task_reference"
                ),
                portfolio_get_all_products_and_their_versions_ref=parameters_to_use.get(
                    "portfolio_get_all_products_and_their_versions_ref"
                ),
                portfolio_get_all_products_and_their_versions_for_hub_ref=parameters_to_use.get(
                    "portfolio_get_all_products_and_their_versions_for_hub_ref"
                ),
            )

    elif section_name == constants.PORTFOLIO_SHARE_AND_ACCEPT_ACCOUNT:
        if status == "terminated":
            raise Exception("Not supported yet")
        else:
            from servicecatalog_puppet.workflow.portfolio.sharing_management import (
                share_and_accept_portfolio_task,
            )

            return (
                share_and_accept_portfolio_task.ShareAndAcceptPortfolioForAccountTask(
                    **common_parameters,
                    portfolio=parameters_to_use.get("portfolio"),
                    share_tag_options=parameters_to_use.get("share_tag_options"),
                    share_principals=parameters_to_use.get("share_principals"),
                    portfolio_task_reference=parameters_to_use.get(
                        "portfolio_task_reference"
                    ),
                    describe_portfolio_shares_task_ref=parameters_to_use.get(
                        "describe_portfolio_shares_task_ref"
                    ),
                )
            )

    elif section_name == constants.PORTFOLIO_SHARE_AND_ACCEPT_AWS_ORGANIZATIONS:
        if status == "terminated":
            raise Exception("Not supported yet")
        else:
            from servicecatalog_puppet.workflow.portfolio.sharing_management import (
                share_portfolio_via_orgs_task,
            )

            return share_portfolio_via_orgs_task.SharePortfolioViaOrgsTask(
                **minimum_common_parameters,
                region=parameters_to_use.get("region"),
                portfolio=parameters_to_use.get("portfolio"),
                share_tag_options=parameters_to_use.get("share_tag_options"),
                share_principals=parameters_to_use.get("share_principals"),
                portfolio_task_reference=parameters_to_use.get(
                    "portfolio_task_reference"
                ),
                ou_to_share_with=parameters_to_use.get("ou_to_share_with"),
                describe_portfolio_shares_task_ref=parameters_to_use.get(
                    "describe_portfolio_shares_task_ref"
                ),
            )

    elif section_name == constants.PORTFOLIO_GET_ALL_PRODUCTS_AND_THEIR_VERSIONS:
        if status == "terminated":
            raise Exception("Not supported yet")
        else:
            from servicecatalog_puppet.workflow.portfolio.accessors import (
                get_all_products_and_their_versions_task,
            )

            return get_all_products_and_their_versions_task.GetAllProductsAndTheirVersionsTask(
                **common_parameters,
                portfolio=parameters_to_use.get("portfolio"),
                portfolio_task_reference=parameters_to_use.get(
                    "portfolio_task_reference"
                ),
            )

    elif section_name == constants.DESCRIBE_PROVISIONING_PARAMETERS:
        if status == "terminated":
            raise Exception("Not supported yet")
        else:
            from servicecatalog_puppet.workflow.launch import (
                provisioning_artifact_parameters_task,
            )

            return provisioning_artifact_parameters_task.ProvisioningArtifactParametersTask(
                **minimum_common_parameters,
                region=parameters_to_use.get("region"),
                portfolio=parameters_to_use.get("portfolio"),
                product=parameters_to_use.get("product"),
                version=parameters_to_use.get("version"),
            )

    elif section_name == constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION:
        if status == "terminated":
            from servicecatalog_puppet.workflow.portfolio.portfolio_management import (
                terminate_associations_task,
            )

            return terminate_associations_task.TerminateAssociationTask(
                **common_parameters,
                portfolio=parameters_to_use.get("portfolio"),
                portfolio_task_reference=parameters_to_use.get(
                    "portfolio_task_reference"
                ),
            )
        else:
            from servicecatalog_puppet.workflow.portfolio.portfolio_management import (
                create_associations_task,
            )

            return create_associations_task.CreateAssociationTask(
                **common_parameters,
                portfolio=parameters_to_use.get("portfolio"),
                portfolio_task_reference=parameters_to_use.get(
                    "portfolio_task_reference"
                ),
            )

    elif section_name == constants.APPS:
        if status == "terminated":
            raise Exception("Not supported yet")
        else:
            from servicecatalog_puppet.workflow.apps import provision_app_task

            return provision_app_task.ProvisionAppTask(
                **common_parameters,
                app_name=parameters_to_use.get("app_name"),
                bucket=parameters_to_use.get("bucket"),
                key=parameters_to_use.get("key"),
                version_id=parameters_to_use.get("version_id"),
                ssm_param_inputs=[],
                launch_parameters=parameters_to_use.get("launch_parameters"),
                manifest_parameters=parameters_to_use.get(""),
                account_parameters=parameters_to_use.get(""),
                retry_count=parameters_to_use.get("retry_count"),
                worker_timeout=parameters_to_use.get("worker_timeout"),
                ssm_param_outputs=[],
                requested_priority=parameters_to_use.get("requested_priority"),
                execution=parameters_to_use.get("execution"),
            )

    elif section_name == constants.WORKSPACES:
        if status == "terminated":
            from servicecatalog_puppet.workflow.workspaces import (
                terminate_workspace_task,
            )

            return terminate_workspace_task.TerminateWorkspaceTask(
                **common_parameters,
                workspace_name=parameters_to_use.get("workspace_name"),
                bucket=parameters_to_use.get("bucket"),
                key=parameters_to_use.get("key"),
                version_id=parameters_to_use.get("version_id"),
                ssm_param_inputs=[],
                launch_parameters=parameters_to_use.get("launch_parameters"),
                manifest_parameters=parameters_to_use.get(""),
                account_parameters=parameters_to_use.get(""),
                retry_count=parameters_to_use.get("retry_count"),
                worker_timeout=parameters_to_use.get("worker_timeout"),
                ssm_param_outputs=[],
                requested_priority=parameters_to_use.get("requested_priority"),
                execution=parameters_to_use.get("execution"),
                manifest_file_path=manifest_file_path,
            )

        else:
            from servicecatalog_puppet.workflow.workspaces import (
                provision_workspace_task,
            )

            return provision_workspace_task.ProvisionWorkspaceTask(
                **common_parameters,
                workspace_name=parameters_to_use.get("workspace_name"),
                bucket=parameters_to_use.get("bucket"),
                key=parameters_to_use.get("key"),
                version_id=parameters_to_use.get("version_id"),
                ssm_param_inputs=[],
                launch_parameters=parameters_to_use.get("launch_parameters"),
                manifest_parameters=parameters_to_use.get(""),
                account_parameters=parameters_to_use.get(""),
                retry_count=parameters_to_use.get("retry_count"),
                worker_timeout=parameters_to_use.get("worker_timeout"),
                ssm_param_outputs=[],
                requested_priority=parameters_to_use.get("requested_priority"),
                execution=parameters_to_use.get("execution"),
                manifest_file_path=manifest_file_path,
            )

    elif section_name == constants.WORKSPACE_ACCOUNT_PREPARATION:
        if status == "terminated":
            raise Exception("Not supported")
        else:
            from servicecatalog_puppet.workflow.workspaces import (
                prepare_account_for_workspace_task,
            )

            return prepare_account_for_workspace_task.PrepareAccountForWorkspaceTask(
                **minimum_common_parameters,
                account_id=parameters_to_use.get("account_id"),
            )

    elif (
        section_name == constants.PORTFOLIO_DISASSOCIATE_ALL_PRODUCTS_AND_THEIR_VERSIONS
    ):
        if status == "terminated":
            raise Exception("Not supported")
        else:
            from servicecatalog_puppet.workflow.portfolio.portfolio_management import (
                disassociate_products_from_portfolio_task,
            )

            return disassociate_products_from_portfolio_task.DisassociateProductsFromPortfolio(
                **common_parameters,
                portfolio=parameters_to_use.get("portfolio"),
                portfolio_task_reference=parameters_to_use.get(
                    "portfolio_task_reference"
                ),
            )

    elif section_name == constants.RUN_DEPLOY_IN_SPOKE:
        from servicecatalog_puppet.workflow.launch import run_deploy_in_spoke_task

        return run_deploy_in_spoke_task.RunDeployInSpokeTask(
            **minimum_common_parameters,
            account_id=parameters_to_use.get("account_id"),
            generate_manifest_ref=parameters_to_use.get("generate_manifest_ref"),
        )

    elif section_name == constants.GENERATE_MANIFEST:
        from servicecatalog_puppet.workflow.manifest import (
            generate_manifest_with_ids_task,
        )

        return generate_manifest_with_ids_task.GenerateManifestWithIdsTask(
            **minimum_common_parameters,
        )

    elif section_name == constants.GET_TEMPLATE_FROM_S3:
        from servicecatalog_puppet.workflow.stack import (
            get_cloud_formation_template_from_s3,
        )

        return get_cloud_formation_template_from_s3.GetCloudFormationTemplateFromS3(
            **common_parameters,
            bucket=parameters_to_use.get("bucket"),
            key=parameters_to_use.get("key"),
            version_id=parameters_to_use.get("version_id"),
        )

    elif section_name == constants.GET_OR_CREATE_SERVICE_CONTROL_POLICIES_POLICY:
        from servicecatalog_puppet.workflow.service_control_policies import (
            get_or_create_policy_task,
        )

        name = parameters_to_use.get("policy_name")
        m = get_m(manifest_file_path)
        tags = m.get(constants.SERVICE_CONTROL_POLICIES).get(name).get("tags", [])

        return get_or_create_policy_task.GetOrCreatePolicyTask(
            **common_parameters,
            policy_name=name,
            policy_description=parameters_to_use.get("policy_description"),
            policy_content=parameters_to_use.get("policy_content"),
            manifest_file_path=manifest_file_path,
            tags=tags,
        )

    elif section_name == constants.GET_OR_CREATE_TAG_POLICIES_POLICY:
        from servicecatalog_puppet.workflow.tag_policies import (
            get_or_create_policy_task,
        )

        name = parameters_to_use.get("policy_name")
        with open(manifest_file_path, "r") as f:
            m = serialisation_utils.load(f.read())
        tags = m.get(constants.TAG_POLICIES).get(name).get("tags", [])

        return get_or_create_policy_task.GetOrCreatePolicyTask(
            **common_parameters,
            policy_name=name,
            policy_description=parameters_to_use.get("policy_description"),
            policy_content=parameters_to_use.get("policy_content"),
            manifest_file_path=manifest_file_path,
            tags=tags,
        )

    elif section_name == constants.PREPARE_ACCOUNT_FOR_STACKS:
        from servicecatalog_puppet.workflow.stack import prepare_account_for_stack_task

        return prepare_account_for_stack_task.PrepareAccountForStackTask(
            **common_parameters,
        )

    elif section_name == constants.CREATE_POLICIES:
        from servicecatalog_puppet.workflow.generate import generate_policies_task

        return generate_policies_task.GeneratePolicies(
            **common_parameters,
            organizations_to_share_with=parameters_to_use.get(
                "organizations_to_share_with"
            ),
            ous_to_share_with=parameters_to_use.get("ous_to_share_with"),
            accounts_to_share_with=parameters_to_use.get("accounts_to_share_with"),
        )

    elif section_name == constants.ORGANIZATIONAL_UNITS:
        from servicecatalog_puppet.workflow.organizational_units import (
            get_or_create_organizational_unit_task,
        )

        return get_or_create_organizational_unit_task.GetOrCreateOrganizationalUnitTask(
            **common_parameters,
            path=parameters_to_use.get("path"),
            parent_ou_id=parameters_to_use.get("parent_ou_id"),
            name=parameters_to_use.get("name"),
            tags=parameters_to_use.get("tags"),
            parent_ou_task_ref=parameters_to_use.get("parent_ou_task_ref"),
        )

    elif section_name == constants.DESCRIBE_PORTFOLIO_SHARES:
        from servicecatalog_puppet.workflow.portfolio.sharing_management import (
            describe_portfolio_shares_task,
        )

        return describe_portfolio_shares_task.DescribePortfolioSharesTask(
            **common_parameters,
            portfolio_task_reference=parameters_to_use.get("portfolio_task_reference"),
            type=parameters_to_use.get("type"),
        )

    elif section_name == constants.C7N_PREPARE_HUB_ACCOUNT_TASK:
        from servicecatalog_puppet.workflow.c7n import prepare_c7n_hub_account_task

        return prepare_c7n_hub_account_task.PrepareC7NHubAccountTask(
            **common_parameters,
            custodian_region=parameters_to_use.get(
                "custodian_region",
            ),
            c7n_version=parameters_to_use.get(
                "c7n_version",
            ),
            c7n_org_version=parameters_to_use.get(
                "c7n_org_version",
            ),
            organization=parameters_to_use.get(
                "organization",
            ),
            role_name=parameters_to_use.get(
                "role_name",
            ),
            role_path=parameters_to_use.get(
                "role_path",
            ),
            schedule_expression=parameters_to_use.get(
                "schedule_expression",
            ),
        )

    elif section_name == constants.C7N_FORWARD_EVENTS_FOR_ACCOUNT_TASK:
        from servicecatalog_puppet.workflow.c7n import forward_events_for_account_tasks

        return forward_events_for_account_tasks.ForwardEventsForAccountTask(
            **common_parameters,
            c7n_account_id=parameters_to_use.get(
                "c7n_account_id",
            ),
            custodian_region=parameters_to_use.get(
                "custodian_region",
            ),
        )

    elif section_name == constants.C7N_FORWARD_EVENTS_FOR_REGION_TASK:
        from servicecatalog_puppet.workflow.c7n import forward_events_for_region_tasks

        return forward_events_for_region_tasks.ForwardEventsForRegionTask(
            **common_parameters,
            c7n_account_id=parameters_to_use.get(
                "c7n_account_id",
            ),
            custodian_region=parameters_to_use.get(
                "custodian_region",
            ),
        )

    elif section_name == constants.C7N_CREATE_CUSTODIAN_ROLE_TASK:
        from servicecatalog_puppet.workflow.c7n import create_custodian_role_task

        return create_custodian_role_task.CreateCustodianRoleTask(
            **common_parameters,
            c7n_account_id=parameters_to_use.get(
                "c7n_account_id",
            ),
            role_name=parameters_to_use.get(
                "role_name",
            ),
            role_path=parameters_to_use.get(
                "role_path",
            ),
            role_managed_policy_arns=parameters_to_use.get(
                "role_managed_policy_arns",
            ),
        )

    elif section_name == constants.C7N_DEPLOY_POLICIES_TASK:
        from servicecatalog_puppet.workflow.c7n import deploy_c7n_policies

        return deploy_c7n_policies.DeployC7NPolicies(
            **common_parameters,
            policies=parameters_to_use.get(
                "policies",
            ),
            deployments=parameters_to_use.get(
                "deployments",
            ),
            role_name=parameters_to_use.get(
                "role_name",
            ),
            role_path=parameters_to_use.get(
                "role_path",
            ),
            uses_orgs=parameters_to_use.get(
                "uses_orgs",
            ),
        )

    elif section_name == constants.TERMINATE_CLOUDFORMATION_STACK_TASK:
        from servicecatalog_puppet.workflow.general import (
            terminate_cloudformation_stack_task,
        )

        return terminate_cloudformation_stack_task.TerminateCloudFormationStackTask(
            **common_parameters,
            stack_name=parameters_to_use.get("stack_name"),
            execution=parameters_to_use.get("execution"),
        )

    else:
        raise Exception(f"Unknown section_name: {section_name}")



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import copy

from deepmerge import always_merger

from servicecatalog_puppet import (
    config,
    constants,
    serialisation_utils,
    task_reference_constants,
)
from servicecatalog_puppet.commands import graph
from servicecatalog_puppet.commands.task_reference_helpers.generators import generator
from servicecatalog_puppet.commands.task_reference_helpers.generators.boto3_parameter_handler import (
    boto3_parameter_handler,
)
from servicecatalog_puppet.commands.task_reference_helpers.generators.s3_parameter_handler import (
    s3_parameter_handler,
)
from servicecatalog_puppet.commands.task_reference_helpers.generators.ssm_outputs_handler import (
    ssm_outputs_handler,
)
from servicecatalog_puppet.commands.task_reference_helpers.generators.ssm_parameter_handler import (
    ssm_parameter_handler,
)
from servicecatalog_puppet.waluigi.shared_tasks.task_topological_generations_without_scheduler_unit_test import (
    dependency_task_reference,
)
from servicecatalog_puppet.workflow import workflow_utils
from servicecatalog_puppet.workflow.dependencies import resources_factory


def generate(puppet_account_id, manifest, output_file_directory_path):
    default_region = config.get_home_region(puppet_account_id)
    regions_in_use = config.get_regions()

    all_tasks = dict()
    tasks_by_type = dict()
    tasks_by_region = dict()
    tasks_by_account_id = dict()
    tasks_by_account_id_and_region = dict()

    #
    # ZERO pass - generate policies
    #
    organizations_to_share_with = dict()
    ous_to_share_with = dict()
    accounts_to_share_with = dict()

    for a in manifest.get("accounts", []):
        if a.get("organization"):
            organizations_to_share_with[a.get("organization")] = True
        if a.get("expanded_from"):
            ous_to_share_with[a.get("expanded_from")] = True
        else:
            accounts_to_share_with[a.get("account_id")] = True

    all_tasks[constants.CREATE_POLICIES] = {
        "execution": constants.EXECUTION_MODE_HUB,
        task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
        task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
        "account_id": puppet_account_id,
        "region": default_region,
        task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
        "task_reference": constants.CREATE_POLICIES,
        "dependencies_by_reference": list(),
        "section_name": constants.CREATE_POLICIES,
        "organizations_to_share_with": list(organizations_to_share_with.keys()),
        "ous_to_share_with": list(ous_to_share_with.keys()),
        "accounts_to_share_with": list(accounts_to_share_with.keys()),
    }

    #
    # First pass - handle tasks
    # First pass - create ssm output tasks
    # First pass - set up spoke local portfolios
    #
    for (
        section_name_singular,
        section_name,
    ) in constants.ALL_SECTION_NAME_SINGULAR_AND_PLURAL_LIST:
        tasks_by_type[section_name_singular] = dict()
        tasks_by_region[section_name_singular] = dict()
        tasks_by_account_id[section_name_singular] = dict()
        tasks_by_account_id_and_region[section_name_singular] = dict()
        for item_name, item in manifest.get(section_name, {}).items():
            tasks_by_type[section_name_singular][item_name] = list()
            tasks_by_region[section_name_singular][item_name] = dict()
            tasks_by_account_id[section_name_singular][item_name] = dict()
            tasks_by_account_id_and_region[section_name_singular][item_name] = dict()
            task_reference_prefix = f"{section_name}_{item_name}"
            tasks_to_add = manifest.get_tasks_for(
                puppet_account_id,
                section_name,
                item_name,
                default_region,
                regions_in_use,
            )
            for task_to_add in tasks_to_add:
                task_to_add[task_reference_constants.MANIFEST_SECTION_NAMES] = {
                    section_name: True
                }
                task_to_add[task_reference_constants.MANIFEST_ITEM_NAMES] = {
                    item_name: True
                }
                task_to_add[task_reference_constants.MANIFEST_ACCOUNT_IDS] = {
                    task_to_add.get("account_id"): True
                }
                task_to_add["section_name"] = section_name
                task_to_add["item_name"] = item_name
                # set up for later pass
                task_to_add["dependencies_by_reference"] = [constants.CREATE_POLICIES]
                task_reference = graph.escape(
                    f"{task_to_add.get('ou_name', task_to_add.get('account_id'))}-{task_to_add.get('region')}"
                )
                all_tasks_task_reference = f"{task_reference_prefix}_{task_reference}"
                task_to_add["task_reference"] = all_tasks_task_reference
                all_tasks[all_tasks_task_reference] = task_to_add
                tasks_by_type[section_name_singular][item_name].append(
                    all_tasks_task_reference
                )

                if not tasks_by_region[section_name_singular][item_name].get(
                    task_to_add.get("region")
                ):
                    tasks_by_region[section_name_singular][item_name][
                        task_to_add.get("region")
                    ] = list()

                tasks_by_region[section_name_singular][item_name][
                    task_to_add.get("region")
                ].append(all_tasks_task_reference)

                if not tasks_by_account_id[section_name_singular][item_name].get(
                    task_to_add.get("account_id")
                ):
                    tasks_by_account_id[section_name_singular][item_name][
                        task_to_add.get("account_id")
                    ] = list()

                tasks_by_account_id[section_name_singular][item_name][
                    task_to_add.get("account_id")
                ].append(all_tasks_task_reference)

                account_and_region = (
                    f'{task_to_add.get("account_id")}-{task_to_add.get("region")}'
                )
                if not tasks_by_account_id_and_region[section_name_singular][
                    item_name
                ].get(account_and_region):
                    tasks_by_account_id_and_region[section_name_singular][item_name][
                        account_and_region
                    ] = list()

                tasks_by_account_id_and_region[section_name_singular][item_name][
                    account_and_region
                ].append(all_tasks_task_reference)

                ssm_outputs_handler(
                    all_tasks,
                    all_tasks_task_reference,
                    default_region,
                    item_name,
                    puppet_account_id,
                    section_name,
                    task_to_add,
                )

                generator.generate(
                    all_tasks,
                    all_tasks_task_reference,
                    item_name,
                    puppet_account_id,
                    section_name,
                    task_reference,
                    task_to_add,
                    manifest,
                )

                task = task_to_add  # TODO rename
                new_tasks = all_tasks  # TODO rename
                parameters = {}
                launch_parameters = (
                    manifest.get(task.get("section_name"), {})
                    .get(task.get("item_name"), {})
                    .get("parameters", {})
                )
                manifest_parameters = copy.deepcopy(manifest.get("parameters"))
                account_parameters = manifest.get_parameters_for_account(
                    task.get("account_id")
                )

                always_merger.merge(parameters, manifest_parameters)
                always_merger.merge(parameters, launch_parameters)
                always_merger.merge(parameters, account_parameters)

                if task.get("status") != constants.TERMINATED:
                    for parameter_name, parameter_details in parameters.items():
                        ssm_parameter_handler(
                            all_tasks,
                            default_region,
                            new_tasks,
                            parameter_details,
                            puppet_account_id,
                            task,
                        )
                        boto3_parameter_handler(
                            new_tasks,
                            parameter_details,
                            parameter_name,
                            puppet_account_id,
                            task,
                        )
                        s3_parameter_handler(
                            all_tasks,
                            default_region,
                            new_tasks,
                            parameter_details,
                            puppet_account_id,
                            task,
                        )

    #
    # Second pass - replacing dependencies with dependencies_by_reference and adding resources
    #
    for task_reference, task in all_tasks.items():
        for dependency in task.get("dependencies", []):
            section = dependency.get("type")
            affinity = dependency.get("affinity")
            name = dependency.get("name")

            if affinity == section:
                if tasks_by_type[section].get(name):
                    task["dependencies_by_reference"].extend(
                        tasks_by_type[section][name]
                    )
                else:
                    # TODO should this be a warning or an error
                    raise Exception(
                        f"invalid use of {affinity} affinity - {name} is not deployed for task {task_reference}"
                    )

            if affinity == constants.AFFINITY_REGION:
                if tasks_by_region[section][name].get(task.get("region")):
                    task["dependencies_by_reference"].extend(
                        tasks_by_region[section][name][task.get("region")]
                    )
                else:
                    # TODO should this be a warning or an error
                    raise Exception(
                        f"invalid use of region affinity - {name} is not deployed in the region: {task.get('region')} for task {task_reference}"
                    )

            if affinity == constants.AFFINITY_ACCOUNT:
                if tasks_by_account_id[section][name].get(task.get("account_id")):
                    task["dependencies_by_reference"].extend(
                        tasks_by_account_id[section][name][task.get("account_id")]
                    )
                else:
                    # TODO should this be a warning or an error
                    raise Exception(
                        f"invalid use of account affinity - {name} is not deployed in the account_id: {task.get('account_id')} for task {task_reference}"
                    )

            if affinity == constants.AFFINITY_ACCOUNT_AND_REGION:
                account_and_region = f"{task.get('account_id')}-{task.get('region')}"
                if tasks_by_account_id_and_region[section][name].get(
                    account_and_region
                ):
                    task["dependencies_by_reference"].extend(
                        tasks_by_account_id_and_region[section][name][
                            account_and_region
                        ]
                    )
                else:
                    # TODO should this be a warning or an error
                    raise Exception(
                        f"invalid use of account-and-region affinity - {name} is not deployed in the account_id and region: {account_and_region} for task {task_reference}"
                    )
        resources = resources_factory.create(
            task.get("section_name"), task, puppet_account_id
        )
        task["resources_required"] = resources

    #
    # Third pass - setting dependencies between parameters and outputs
    #
    for task_reference, task in all_tasks.items():
        # wire up dependencies for SSM parameters to outputs
        ssm_parameters = task.get("ssm_parameters_tasks_references", {}).items()
        dependencies_to_add = []
        for parameter_name, parameter_task_reference in ssm_parameters:
            for dependency in task.get("dependencies_by_reference"):
                ssm_outputs = (
                    all_tasks.get(dependency)
                    .get("ssm_outputs_tasks_references", {})
                    .items()
                )
                for output_name, output_task_reference in ssm_outputs:
                    if (
                        output_task_reference.replace(
                            constants.SSM_OUTPUTS, constants.SSM_PARAMETERS
                        )
                        == parameter_task_reference
                    ):
                        dependencies_to_add.append(output_task_reference)
                        all_tasks[parameter_task_reference][
                            "dependencies_by_reference"
                        ].append(output_task_reference)
        task["dependencies_by_reference"].extend(dependencies_to_add)

        # wire up dependencies for boto3 parameters
        boto3_parameters = task.get("boto3_parameters_tasks_references", {}).items()
        for parameter_name, parameter_task_reference in boto3_parameters:
            parameter_task = all_tasks.get(parameter_task_reference)
            for dependency_task_reference in task.get("dependencies_by_reference"):
                if dependency_task_reference not in parameter_task.get(
                    "dependencies_by_reference"
                ):
                    if dependency_task_reference != parameter_task_reference:
                        if (
                            all_tasks[dependency_task_reference]["section_name"]
                            in constants.ALL_SECTION_NAMES
                        ):
                            parameter_task["dependencies_by_reference"].append(
                                dependency_task_reference
                            )

            if task_reference in parameter_task["dependencies_by_reference"]:
                parameter_task["dependencies_by_reference"].remove(task_reference)

        # wire up dependencies for s3 parameters
        s3_parameters = task.get("s3_object_task_reference", {}).items()
        for parameter_name, parameter_task_reference in s3_parameters:
            parameter_task = all_tasks.get(parameter_task_reference)
            for dependency_task_reference in task.get("dependencies_by_reference"):
                if dependency_task_reference not in parameter_task.get(
                    "dependencies_by_reference"
                ):
                    if dependency_task_reference != parameter_task_reference:
                        if (
                            all_tasks[dependency_task_reference]["section_name"]
                            in constants.ALL_SECTION_NAMES
                        ):
                            parameter_task["dependencies_by_reference"].append(
                                dependency_task_reference
                            )

            if task_reference in parameter_task["dependencies_by_reference"]:
                parameter_task["dependencies_by_reference"].remove(task_reference)

    #
    # Fourth pass - removing cyclic dependencies caused by a->b->c when using boto3 parameters
    # setting up deploy in spoke collections
    #
    accounts_output = dict()
    for task_reference, task in all_tasks.items():
        # remove dependencies on self
        boto3_parameters = task.get("boto3_parameters_tasks_references", {}).items()
        dependencies_to_add = []
        for parameter_name, parameter_task_reference in boto3_parameters:
            parameter_task = all_tasks.get(parameter_task_reference)
            if task_reference in parameter_task.get("dependencies_by_reference"):
                parameter_task["dependencies_by_reference"].remove(task_reference)

        account_id = task.get("account_id")
        if accounts_output.get(account_id) is None:
            accounts_output[account_id] = dict()
        accounts_output[account_id][task.get("task_reference")] = task

    # write the tasks to disk
    for account_id, tasks in accounts_output.items():
        output_file_path = (
            f"{output_file_directory_path}/manifest-task-reference-{account_id}.json"
        )
        open(output_file_path, "w").write(
            serialisation_utils.dump_as_json(dict(all_tasks=tasks))
        )

    reference = dict(
        all_tasks=all_tasks,
    )
    workflow_utils.ensure_no_cyclic_dependencies("complete task reference", all_tasks)
    output_file_path = output_file_directory_path + "/manifest-task-reference-full.json"
    open(output_file_path, "w").write(serialisation_utils.dump_as_json(reference))
    return reference



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import functools
import logging

import yaml
from betterboto import client as betterboto_client

from servicecatalog_puppet import config, constants


logger = logging.getLogger()


@functools.lru_cache(maxsize=32)
def get_config(puppet_account_id, default_region=None):
    logger.info("getting config,  default_region: {}".format(default_region))
    region = (
        default_region if default_region else config.get_home_region(puppet_account_id)
    )
    with betterboto_client.CrossAccountClientContextManager(
        "ssm",
        config.get_puppet_role_arn(puppet_account_id),
        f"{puppet_account_id}-{region}-{config.get_puppet_role_name()}",
        region_name=region,
    ) as ssm:
        response = ssm.get_parameter(Name=constants.CONFIG_PARAM_NAME)
        return yaml.safe_load(response.get("Parameter").get("Value"))


@functools.lru_cache(maxsize=32)
def get_should_use_sns(puppet_account_id, default_region=None):
    logger.info(
        f"getting {constants.CONFIG_SHOULD_COLLECT_CLOUDFORMATION_EVENTS},  default_region: {default_region}"
    )
    return get_config(puppet_account_id, default_region).get(
        constants.CONFIG_SHOULD_COLLECT_CLOUDFORMATION_EVENTS, True
    )


@functools.lru_cache(maxsize=32)
def get_should_delete_rollback_complete_stacks(puppet_account_id, default_region=None):
    logger.info(
        f"getting {constants.CONFIG_SHOULD_DELETE_ROLLBACK_COMPLETE_STACKS},  default_region: {default_region}"
    )
    return get_config(puppet_account_id, default_region).get(
        constants.CONFIG_SHOULD_DELETE_ROLLBACK_COMPLETE_STACKS,
        constants.CONFIG_SHOULD_DELETE_ROLLBACK_COMPLETE_STACKS_DEFAULT,
    )


@functools.lru_cache(maxsize=32)
def get_should_use_product_plans(puppet_account_id, default_region=None):
    logger.info(
        "getting should_use_product_plans,  default_region: {}".format(default_region)
    )
    return get_config(puppet_account_id, default_region).get(
        "should_use_product_plans", True
    )


@functools.lru_cache(maxsize=32)
def get_initialiser_stack_tags():
    with betterboto_client.ClientContextManager("ssm") as ssm:
        try:
            response = ssm.get_parameter(
                Name=constants.INITIALISER_STACK_NAME_SSM_PARAMETER
            )
            initialiser_stack_name = response.get("Parameter").get("Value")
            with betterboto_client.ClientContextManager(
                "cloudformation"
            ) as cloudformation:
                paginator = cloudformation.get_paginator("describe_stacks")
                for page in paginator.paginate(
                    StackName=initialiser_stack_name,
                ):
                    for stack in page.get("Stacks", []):
                        if stack.get("StackStatus") in [
                            "CREATE_IN_PROGRESS",
                            "CREATE_COMPLETE",
                            "UPDATE_IN_PROGRESS",
                            "UPDATE_COMPLETE_CLEANUP_IN_PROGRESS",
                            "UPDATE_COMPLETE",
                        ]:
                            return stack.get("Tags")
                        else:
                            raise Exception(
                                f"Initialiser stack: {initialiser_stack_name} in state: {stack.get('StackStatus')}"
                            )
        except ssm.exceptions.ParameterNotFound:
            logger.warning(
                f"Could not find SSM parameter: {constants.INITIALISER_STACK_NAME_SSM_PARAMETER}, do not know the tags to use"
            )
            return []


@functools.lru_cache(maxsize=32)
def get_should_use_eventbridge(puppet_account_id, default_region=None):
    logger.info(
        "getting should_use_eventbridge,  default_region: {}".format(default_region)
    )
    return get_config(puppet_account_id, default_region).get(
        "should_forward_events_to_eventbridge", False
    )


@functools.lru_cache(maxsize=32)
def get_should_forward_failures_to_opscenter(puppet_account_id, default_region=None):
    logger.info(
        "getting should_forward_failures_to_opscenter,  default_region: {}".format(
            default_region
        )
    )
    return get_config(puppet_account_id, default_region).get(
        "should_forward_failures_to_opscenter", False
    )


@functools.lru_cache(maxsize=32)
def get_drift_token(puppet_account_id, default_region=None):
    logger.info("getting drift_token,  default_region: {}".format(default_region))
    return get_config(puppet_account_id, default_region).get("drift_token", None)


@functools.lru_cache(maxsize=32)
def get_regions(puppet_account_id, default_region=None):
    logger.info(
        f"getting {constants.CONFIG_REGIONS},  default_region: {default_region}"
    )
    return get_config(puppet_account_id, default_region).get(constants.CONFIG_REGIONS)


@functools.lru_cache(maxsize=32)
def get_puppet_account_id():
    with betterboto_client.ClientContextManager("sts") as sts:
        return sts.get_caller_identity().get("Account")


@functools.lru_cache(maxsize=32)
def get_current_account_id():
    with betterboto_client.ClientContextManager("sts") as sts:
        return sts.get_caller_identity().get("Account")


@functools.lru_cache(maxsize=32)
def is_caching_enabled(puppet_account_id, default_region=None):
    return get_config(puppet_account_id, default_region).get(
        "is_caching_enabled", False
    )


@functools.lru_cache(maxsize=32)
def get_global_sharing_mode_default(puppet_account_id, default_region=None):
    logger.info(
        "getting global_sharing_mode_default,  default_region: {}".format(
            default_region
        )
    )
    return get_config(puppet_account_id, default_region).get(
        "global_sharing_mode_default", constants.SHARING_MODE_DEFAULT
    )


@functools.lru_cache(maxsize=32)
def get_global_share_tag_options_default(puppet_account_id, default_region=None):
    logger.info(
        "getting global_share_tag_options_default,  default_region: {}".format(
            default_region
        )
    )
    return get_config(puppet_account_id, default_region).get(
        "global_share_tag_options_default", constants.SHARE_TAG_OPTIONS_DEFAULT
    )


@functools.lru_cache(maxsize=32)
def get_global_share_principals_default(puppet_account_id, default_region=None):
    logger.info(
        "getting global_share_principals_default,  default_region: {}".format(
            default_region
        )
    )
    return get_config(puppet_account_id, default_region).get(
        "global_share_principals_default", constants.SHARE_PRINCIPALS_DEFAULT
    )


@functools.lru_cache(maxsize=32)
def get_aws_sts_regional_endpoints(puppet_account_id, default_region=None):
    logger.info(
        "getting aws_sts_regional_endpoints,  default_region: {}".format(default_region)
    )
    return get_config(puppet_account_id, default_region).get(
        "aws_sts_regional_endpoints", constants.AWS_STS_REGIONAL_ENDPOINTS_DEFAULT
    )


def get_spoke_deploy_environment_compute_type(puppet_account_id, default_region):
    logger.info(
        "getting spoke_deploy_environment_compute_type,  default_region: {}".format(
            default_region
        )
    )
    return get_config(puppet_account_id, default_region).get(
        "spoke_deploy_environment_compute_type",
        constants.SPOKE_EXECUTION_MODE_DEPLOY_ENV_DEFAULT,
    )


def get_scheduler_threads_or_processes(puppet_account_id, default_region):
    logger.info(
        "getting scheduler_threads_or_processes,  default_region: {}".format(
            default_region
        )
    )
    return get_config(puppet_account_id, default_region).get(
        "scheduler_threads_or_processes",
        constants.SCHEDULER_THREADS_OR_PROCESSES_DEFAULT,
    )


def get_scheduler_algorithm(puppet_account_id, default_region):
    logger.info(
        "getting scheduler_algorithm,  default_region: {}".format(default_region)
    )
    return get_config(puppet_account_id, default_region).get(
        "scheduler_algorithm",
        constants.SCHEDULER_ALGORITHM_DEFAULT,
    )



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
import unittest
from copy import deepcopy
from unittest.mock import MagicMock

from servicecatalog_puppet import constants


class TestManifestForExpand(unittest.TestCase):

    organization_id = "o-aaaaaaaa"

    account_a = {
        "account_id": "012345678910",
        "default_region": "eu-west-1",
        "name": "accounta",
        "regions_enabled": [
            "eu-west-2",
        ],
        "tags": ["group:A"],
    }
    account_b = {
        "account_id": "009876543210",
        "default_region": "us-west-1",
        "name": "accountb",
        "regions_enabled": [
            "us-west-2",
        ],
        "tags": ["group:B"],
    }
    account_c = {
        "account_id": "432100098765",
        "default_region": "ap-west-1",
        "name": "accountc",
        "regions_enabled": [
            "ap-west-2",
        ],
        "tags": ["group:C"],
    }
    accounts = [
        account_a,
        account_b,
        account_c,
    ]

    account_ou_a = {
        "ou": "ou-aaaa-aaaaaaaa",
        "default_region": "eu-west-1",
        "name": "oua",
        "regions_enabled": [
            "eu-west-2",
        ],
        "tags": ["group:A"],
    }
    account_ou_b = {
        "ou": "/OrgUnitB",
        "ou_name": "/OrgUnitB",
        "default_region": "us-west-1",
        "name": "oub",
        "regions_enabled": [
            "us-west-2",
        ],
        "tags": ["group:B"],
    }
    account_ou_c = {
        "ou": "ou-aaaa-cccccccc",
        "default_region": "ap-west-1",
        "name": "ouc",
        "regions_enabled": [
            "ap-west-2",
        ],
        "tags": ["group:C"],
    }

    org_units = [
        account_ou_a,
        account_ou_b,
        account_ou_c,
    ]

    def setUp(self):
        self.maxDiff = None
        from servicecatalog_puppet import manifest_utils

        self.client_mock = MagicMock()

        self.sut = manifest_utils.expand_manifest

    def tearDown(self):
        self.sut = None
        self.client_mock = None

    def test_expand_for_single_account_in_ou(self):
        # setup
        expanded_manifest = dict()
        expanded_manifest[constants.ACCOUNTS] = self.accounts
        expanded_manifest[constants.LAUNCHES] = dict()
        expanded_manifest[constants.STACKS] = dict()
        expanded_manifest[constants.SPOKE_LOCAL_PORTFOLIOS] = dict()
        expanded_manifest[constants.ACTIONS] = dict()
        expanded_manifest[constants.ASSERTIONS] = dict()
        expanded_manifest[constants.LAMBDA_INVOCATIONS] = dict()

        expected_results = deepcopy(expanded_manifest)
        for account in expected_results[constants.ACCOUNTS]:
            account_id = account["account_id"]
            account["email"] = f"{account_id}@test.com"
            account["organization"] = self.organization_id

        def describe_account_side_effect(AccountId):
            return {
                "Account": {
                    "Id": AccountId,
                    "Arn": f"arn:aws:organizations::000000000000:account/{self.organization_id}/{AccountId}",
                    "Email": f"{AccountId}@test.com",
                    "Status": "ACTIVE",
                }
            }

        self.client_mock.describe_account = MagicMock(
            side_effect=describe_account_side_effect
        )

        # exercise
        _, actual_results = self.sut(expanded_manifest, self.client_mock)

        # verify
        self.assertDictEqual(expected_results, actual_results)

    def test_expand_for_ou(self):
        # setup
        expanded_manifest = dict()
        expanded_manifest[constants.ACCOUNTS] = self.org_units
        expanded_manifest[constants.LAUNCHES] = dict()
        expanded_manifest[constants.STACKS] = dict()
        expanded_manifest[constants.SPOKE_LOCAL_PORTFOLIOS] = dict()
        expanded_manifest[constants.ACTIONS] = dict()
        expanded_manifest[constants.ASSERTIONS] = dict()
        expanded_manifest[constants.LAMBDA_INVOCATIONS] = dict()

        expected_results = deepcopy(expanded_manifest)
        accounts = []

        account_a = deepcopy(self.account_a)
        account_a["email"] = f"{account_a['account_id']}@test.com"
        account_a["organization"] = self.organization_id
        account_a["expanded_from"] = "ou-aaaa-aaaaaaaa"
        account_a["name"] = account_a["account_id"]
        accounts.append(account_a)

        account_b = deepcopy(self.account_b)
        account_b["email"] = f"{account_b['account_id']}@test.com"
        account_b["organization"] = self.organization_id
        account_b["expanded_from"] = "ou-aaaa-bbbbbbbb"
        account_b["ou_name"] = "/OrgUnitB"
        account_b["name"] = account_b["account_id"]
        accounts.append(account_b)

        account_c = deepcopy(self.account_c)
        account_c["email"] = f"{account_c['account_id']}@test.com"
        account_c["organization"] = self.organization_id
        account_c["expanded_from"] = "ou-aaaa-cccccccc"
        account_c["name"] = account_c["account_id"]
        accounts.append(account_c)

        expected_results[constants.ACCOUNTS] = accounts

        def describe_account_side_effect(AccountId):
            return {
                "Account": {
                    "Id": AccountId,
                    "Arn": f"arn:aws:organizations::000000000000:account/{self.organization_id}/{AccountId}",
                    "Email": f"{AccountId}@test.com",
                    "Status": "ACTIVE",
                    "Name": AccountId,
                }
            }

        def list_children_nested_side_effect(ParentId, ChildType):
            ou_mapping = {
                "ou-aaaa-aaaaaaaa": [{"Id": "012345678910"}],
                "ou-aaaa-bbbbbbbb": [{"Id": "009876543210"}],
                "ou-aaaa-cccccccc": [{"Id": "432100098765"}],
            }
            return ou_mapping.get(ParentId, [])

        self.client_mock.describe_account = MagicMock(
            side_effect=describe_account_side_effect
        )
        self.client_mock.list_children_nested = MagicMock(
            side_effect=list_children_nested_side_effect
        )
        self.client_mock.convert_path_to_ou = MagicMock(return_value="ou-aaaa-bbbbbbbb")

        # exercise
        _, actual_results = self.sut(expanded_manifest, self.client_mock)

        # verify
        self.assertDictEqual(expected_results, actual_results)



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import os
import queue
import time

from betterboto import client as betterboto_client

from servicecatalog_puppet import config, constants
from servicecatalog_puppet.waluigi.dag_utils import logger


def on_task_processing_time_task(
    task_processing_time_queue, complete_event, execution_mode
):
    if execution_mode != constants.EXECUTION_MODE_SPOKE:
        with betterboto_client.CrossAccountClientContextManager(
            "cloudwatch",
            config.get_reporting_role_arn(config.get_executor_account_id()),
            "cloudwatch-puppethub",
        ) as cloudwatch:
            while not complete_event.is_set():
                time.sleep(0.1)
                try:
                    duration, task_type, task_params = task_processing_time_queue.get(
                        timeout=5
                    )
                except queue.Empty:
                    continue
                else:
                    dimensions = [
                        dict(
                            Name="task_type",
                            Value=task_type,
                        ),
                        dict(
                            Name="codebuild_build_id",
                            Value=os.getenv("CODEBUILD_BUILD_ID", "LOCAL_BUILD"),
                        ),
                    ]
                    for note_worthy in [
                        "launch_name",
                        "region",
                        "account_id",
                        "puppet_account_id",
                        "portfolio",
                        "product",
                        "version",
                    ]:
                        if task_params.get(note_worthy):
                            dimensions.append(
                                dict(
                                    Name=str(note_worthy),
                                    Value=str(task_params.get(note_worthy)),
                                )
                            )
                    cloudwatch.put_metric_data(
                        Namespace=f"ServiceCatalogTools/Puppet/v2/ProcessingTime/Tasks",
                        MetricData=[
                            dict(
                                MetricName="Tasks",
                                Dimensions=[dict(Name="TaskType", Value=task_type)]
                                + dimensions,
                                Value=duration,
                                Unit="Seconds",
                            ),
                        ],
                    )
        logger.info("shutting down")
